# Copyright (c) 2008 Chris Moyer http://coredumped.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import urllib
import xml.sax
import uuid
import boto
import boto.utils
import urllib
from boto import handler
from boto.connection import AWSQueryConnection
from boto.resultset import ResultSet
from boto.exception import FPSResponseError

class FPSConnection(AWSQueryConnection):

	APIVersion = '2007-01-08'
	SignatureVersion = '1'

	def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
						is_secure=True, port=None, proxy=None, proxy_port=None,
						host='fps.sandbox.amazonaws.com', debug=0,
						https_connection_factory=None):
		AWSQueryConnection.__init__(self, aws_access_key_id,
												aws_secret_access_key,
												is_secure, port, proxy, proxy_port,
												host, debug, https_connection_factory)
	
	def install_payment_instruction(self, instruction, token_type="Unrestricted", transaction_id=None):
		"""
		InstallPaymentInstruction
		instruction: The PaymentInstruction to send, for example: 
			MyRole=='Caller' orSay 'Roles do not match';
		token_type: Defaults to "Unrestricted"
		transaction_id: Defaults to a new ID
		"""

		if(transaction_id == None):
			transaction_id = uuid.uuid4()
		params = {}
		params['PaymentInstruction'] = instruction
		params['TokenType'] = token_type
		params['CallerReference'] = transaction_id
		response = self.make_request("InstallPaymentInstruction", params)
		return response
	
	def install_caller_instruction(self, token_type="Unrestricted", transaction_id=None):
		"""
		Set us up as a caller
		This will install a new caller_token into the FPS section.
		This should really only be called to regenerate the caller token.
		"""
		response = self.install_payment_instruction("MyRole=='Caller';", token_type=token_type, transaction_id=transaction_id)
		body = response.read()
		if(response.status == 200):
			rs = ResultSet()
			h = handler.XmlHandler(rs, self)
			xml.sax.parseString(body, h)
			caller_token = rs.TokenId
			try:
				boto.config.save_system_option("FPS", "caller_token", caller_token)
			except(IOError):
				boto.config.save_user_option("FPS", "caller_token", caller_token)
			return caller_token
		else:
			raise FPSResponseError(response.status, respons.reason, body)

	def install_recipient_instruction(self, token_type="Unrestricted", transaction_id=None):
		"""
		Set us up as a Recipient
		This will install a new caller_token into the FPS section.
		This should really only be called to regenerate the recipient token.
		"""
		response = self.install_payment_instruction("MyRole=='Recipient';", token_type=token_type, transaction_id=transaction_id)
		body = response.read()
		if(response.status == 200):
			rs = ResultSet()
			h = handler.XmlHandler(rs, self)
			xml.sax.parseString(body, h)
			recipient_token = rs.TokenId
			try:
				boto.config.save_system_option("FPS", "recipient_token", recipient_token)
			except(IOError):
				boto.config.save_user_option("FPS", "recipient_token", recipient_token)

			return recipient_token
		else:
			raise FPSResponseError(response.status, respons.reason, body)

	def make_url(self, returnURL, paymentReason, pipelineName, **params):
		"""
		Generate the URL with the signature required for a transaction
		"""
		params['callerKey'] = str(self.aws_access_key_id)
		params['returnURL'] = str(returnURL)
		params['paymentReason'] = str(paymentReason)
		params['pipelineName'] = pipelineName

		if(not params.has_key('callerReference')):
			params['callerReference'] = str(uuid.uuid4())

		url = ""
		keys = params.keys()
		keys.sort()
		for k in keys:
			url += "&%s=%s" % (k, urllib.quote_plus(str(params[k])))

		url = "/cobranded-ui/actions/start?%s" % ( url[1:])
		signature= boto.utils.encode(self.aws_secret_access_key, url, True)
		return "https://authorize.payments-sandbox.amazon.com%s&awsSignature=%s" % (url, signature)

	def make_payment(self, amount, sender_token, charge_fee_to="Recipient", reference=None, senderReference=None, recipientReference=None, senderDescription=None, recipientDescription=None, callerDescription=None, metadata=None, transactionDate=None):
		"""
		Make a payment transaction
		You must specify the amount and the sender token.
		"""
		params = {}
		params['RecipientTokenId'] = boto.config.get("FPS", "recipient_token")
		params['CallerTokenId'] = boto.config.get("FPS", "caller_token")
		params['SenderTokenId'] = sender_token
		params['TransactionAmount.Amount'] = str(amount)
		params['TransactionAmount.CurrencyCode'] = "USD"
		params['ChargeFeeTo'] = charge_fee_to

		if(transactionDate != None):
			params['TransactionDate'] = transactionDate
		if(senderReference != None):
			params['SenderReference'] = senderReference
		if(recipientReference != None):
			params['RecipientReference'] = recipientReference
		if(senderDescription != None):
			params['SenderDescription'] = senderDescription
		if(recipientDescription != None):
			params['RecipientDescription'] = recipientDescription
		if(callerDescription != None):
			params['CallerDescription'] = callerDescription
		if(metadata != None):
			params['MetaData'] = metadata
		if(transactionDate != None):
			params['TransactionDate'] = transactionDate
		if(reference == None):
			reference = uuid.uuid4()
		params['CallerReference'] = reference

		response = self.make_request("Pay", params)
		body = response.read()
		if(response.status == 200):
			rs = ResultSet()
			h = handler.XmlHandler(rs, self)
			xml.sax.parseString(body, h)
			return rs
		else:
			raise FPSResponseError(response.status, response.reason, body)



# Copyright (c) 2006,2007,2008 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
from boto.pyami.installers.ubuntu.installer import Installer
import os
import boto
from ConfigParser import SafeConfigParser
import subprocess
import time

class MySQL(Installer):

    def install(self):
        self.run('apt-get update')
        self.run('apt-get -y install mysql-server', notify=True, exit_on_error=True)

    def set_root_password(self, password=None):
        if not password:
            password = boto.config.get('Pyami', 'mysql_root_password')
        if password:
            self.run('mysqladmin -u root password %s' % password)

    def change_data_dir(self):
        fresh_install = False;
        time.sleep(10) #trying to stop mysql immediately after installing it fails
        # We need to wait until mysql creates the root account before we kill it
        # or bad things will happen
        while self.run("echo 'quit' | mysql -u root") != 0:
            time.sleep(5)
        self.run('/etc/init.d/mysql stop')
        self.run("pkill -9 mysql")

        if not os.path.exists('/mnt/mysql'):
            self.run('mkdir /mnt/mysql')
            fresh_install = True;
        self.run('chown -R mysql:mysql /mnt/mysql')
        fp = open('/etc/mysql/conf.d/use_mnt.cnf', 'w')
        fp.write('# created by pyami\n')
        fp.write('# use the /mnt volume for data\n')
        fp.write('[mysqld]\n')
        fp.write('datadir = /mnt/mysql\n')
        fp.write('log_bin = /mnt/mysql/mysql-bin.log\n')
        fp.close()
        if fresh_install:
            self.run('cp -pr /var/lib/mysql/* /mnt/mysql/')
            self.start('mysql')
        else:
            #get the password ubuntu expects to use:
            config_parser = SafeConfigParser()
            config_parser.read('/etc/mysql/debian.cnf')
            password = config_parser.get('client', 'password')
            # start the mysql deamon, then mysql with the required grant statement piped into it:
            self.start('mysql')
            time.sleep(10) #time for mysql to start
            grant_command = "echo \"GRANT ALL PRIVILEGES ON *.* TO 'debian-sys-maint'@'localhost' IDENTIFIED BY '%s' WITH GRANT OPTION;\" | mysql" % password
            while self.run(grant_command) != 0:
                time.sleep(5)
            # leave mysqld running

    def main(self):
        self.install()
        self.set_root_password()
        self.change_data_dir()
        



# Copyright (c) 2006-2008 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

import random, time, os, datetime
import boto
from boto.sdb.persist.object import SDBObject
from boto.sdb.persist.property import *

class Identifier(object):

    _hex_digits = '0123456789abcdef'

    @classmethod
    def gen(cls, prefix):
        suffix = ''
        for i in range(0,8):
            suffix += random.choice(cls._hex_digits)
        return ts + '-' + suffix
    
class Version(SDBObject):

    name = StringProperty()
    pdb = ObjectProperty(ref_class=SDBObject)
    date = DateTimeProperty()
    
    def __init__(self, id=None, manager=None):
        SDBObject.__init__(self, id, manager)
        if id == None:
            self.name = Identifier.gen('v')
            self.date = datetime.datetime.now()
            print 'created Version %s' % self.name

    def partitions(self):
        """
        Return an iterator containing all Partition objects related to this Version.

        @rtype: iterator of L{Partitions<boto.mapreduce.partitiondb.Partition>}
        @return: The Partitions in this Version
        """
        return self.get_related_objects('version', Partition)

    def add_partition(self, name=None):
        """
        Add a new Partition to this Version.

        @type name: string
        @param name: The name of the new Partition (optional)

        @rtype: L{Partition<boto.mapreduce.partitiondb.Partition>}
        @return: The new Partition object
        """
        p = Partition(manager=self.manager, name=name)
        p.version = self
        p.pdb = self.pdb
        p.save()
        return p

    def get_s3_prefix(self):
        if not self.pdb:
            raise ValueError, 'pdb attribute must be set to compute S3 prefix'
        return self.pdb.get_s3_prefix() + self.name + '/'

class PartitionDB(SDBObject):

    name = StringProperty()
    bucket_name = StringProperty()
    versions = ObjectListProperty(ref_class=Version)

    def __init__(self, id=None, manager=None, name='', bucket_name=''):
        SDBObject.__init__(self, id, manager)
        if id == None:
            self.name = name
            self.bucket_name = bucket_name

    def get_s3_prefix(self):
        return self.name + '/'

    def add_version(self):
        """
        Add a new Version to this PartitionDB.  The newly added version becomes the
        current version.

        @rtype: L{Version<boto.mapreduce.partitiondb.Version>}
        @return: The newly created Version object.
        """
        v = Version()
        v.pdb = self
        v.save()
        self.versions.append(v)
        return v

    def revert(self):
        """
        Revert to the previous version of this PartitionDB.  The current version is removed from the
        list of Versions and the Version immediately preceeding it becomes the current version.
        Note that this method does not delete the Version object or any Partitions related to the
        Version object.

        @rtype: L{Version<boto.mapreduce.partitiondb.Version>}
        @return: The previous current Version object.
        """
        v = self.current_version()
        if v:
            self.versions.remove(v)
        return v

    def current_version(self):
        """
        Get the currently active Version of this PartitionDB object.

        @rtype: L{Version<boto.mapreduce.partitiondb.Version>}
        @return: The current Version object or None if there are no Versions associated
                 with this PartitionDB object.
        """
        if self.versions:
            if len(self.versions) > 0:
                return self.versions[-1]
        return None

class Partition(SDBObject):

    def __init__(self, id=None, manager=None, name=None):
        SDBObject.__init__(self, id, manager)
        if id == None:
            self.name = name

    name = StringProperty()
    version = ObjectProperty(ref_class=Version)
    pdb = ObjectProperty(ref_class=PartitionDB)
    data = S3KeyProperty()

    def get_key_name(self):
        return self.version.get_s3_prefix() + self.name

    def upload(self, path, bucket_name=None):
        if not bucket_name:
            bucket_name = self.version.pdb.bucket_name
        s3 = self.manager.get_s3_connection()
        bucket = s3.lookup(bucket_name)
        directory, filename = os.path.split(path)
        self.name = filename
        key = bucket.new_key(self.get_key_name())
        key.set_contents_from_filename(path)
        self.data = key
        self.save()

    def delete(self):
        if self.data:
            self.data.delete()
        SDBObject.delete(self)
        
        
        



# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import mimetypes
import os
import rfc822
import StringIO
import base64
import boto.utils
from boto.exception import S3ResponseError, S3DataError, BotoClientError
from boto.s3.user import User
from boto import UserAgent, config
try:
    from hashlib import md5
except ImportError:
    from md5 import md5

class Key:

    DefaultContentType = 'application/octet-stream'

    BufferSize = 8192

    def __init__(self, bucket=None, name=None):
        self.bucket = bucket
        self.name = name
        self.metadata = {}
        self.content_type = self.DefaultContentType
        self.content_encoding = None
        self.filename = None
        self.etag = None
        self.last_modified = None
        self.owner = None
        self.storage_class = None
        self.md5 = None
        self.base64md5 = None
        self.path = None
        self.resp = None
        self.mode = None
        self.size = None

    def __repr__(self):
        if self.bucket:
            return '<Key: %s,%s>' % (self.bucket.name, self.name)
        else:
            return '<Key: None,%s>' % self.name

    def __getattr__(self, name):
        if name == 'key':
            return self.name
        else:
            raise AttributeError

    def __setattr__(self, name, value):
        if name == 'key':
            self.__dict__['name'] = value
        else:
            self.__dict__[name] = value

    def __iter__(self):
        return self

    def open_read(self, headers=None, query_args=None):
        """
        Open this key for reading
        
        @type headers: dict
        @param headers: Headers to pass in the web request
        
        @type query_args: string
        @param query_args: Arguments to pass in the query string (ie, 'torrent')
        """
        if self.resp == None:
            self.mode = 'r'
            
            self.resp = self.bucket.connection.make_request('GET', self.bucket.name, self.name, headers, query_args=query_args)
            if self.resp.status < 199 or self.resp.status > 299:
                raise S3ResponseError(self.resp.status, self.resp.reason)
            response_headers = self.resp.msg
            self.metadata = boto.utils.get_aws_metadata(response_headers)
            for name,value in response_headers.items():
                if name.lower() == 'content-length':
                    self.size = int(value)
                elif name.lower() == 'etag':
                    self.etag = value
                elif name.lower() == 'content-type':
                    self.content_type = value
                elif name.lower() == 'content-encoding':
                    self.content_encoding = value
                elif name.lower() == 'last-modified':
                    self.last_modified = value

    def open_write(self, headers=None):
        """
        Open this key for writing. 
        Not yet implemented
        
        @type headers: dict
        @param headers: Headers to pass in the write request
        """
        raise BotoClientError('Not Implemented')

    def open(self, mode='r', headers=None, query_args=None):
        if mode == 'r':
            self.mode = 'r'
            self.open_read(headers=headers, query_args=query_args)
        elif mode == 'w':
            self.mode = 'w'
            self.open_write(headers=headers)
        else:
            raise BotoClientError('Invalid mode: %s' % mode)

    closed = False
    def close(self):
        if self.resp:
            self.resp.read()
        self.resp = None
        self.mode = None
        self.closed = True
    
    def next(self):
        """
        By providing a next method, the key object supports use as an iterator.
        For example, you can now say:

        for bytes in key:
            write bytes to a file or whatever

        All of the HTTP connection stuff is handled for you.
        """
        self.open_read()
        data = self.resp.read(self.BufferSize)
        if not data:
            self.close()
            raise StopIteration
        return data

    def read(self, size=0):
        if size == 0:
            size = self.BufferSize
        self.open_read()
        data = self.resp.read(size)
        if not data:
            self.close()
        return data

    def copy(self, dst_bucket, dst_key, metadata=None):
        """
        Copy this Key to another bucket.

        @type dst_bucket: string
        @param dst_bucket: The name of the destination bucket

        @type dst_key: string
        @param dst_key: The name of the destinatino key
        
        @type metadata: dict
        @param metadata: Metadata to be associated with new key.
                         If metadata is supplied, it will replace the
                         metadata of the source key being copied.
                         If no metadata is supplied, the source key's
                         metadata will be copied to the new key.

        @rtype: L{Key<boto.s3.key.Key>} or subclass
        @returns: An instance of the newly created key object
        """
        dst_bucket = self.bucket.connection.lookup(dst_bucket)
        return dst_bucket.copy_key(dst_key, self.bucket.name, self.name, metadata)

    def startElement(self, name, attrs, connection):
        if name == 'Owner':
            self.owner = User(self)
            return self.owner
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'Key':
            self.name = value.encode('utf-8')
        elif name == 'ETag':
            self.etag = value
        elif name == 'LastModified':
            self.last_modified = value
        elif name == 'Size':
            self.size = int(value)
        elif name == 'StorageClass':
            self.storage_class = value
        elif name == 'Owner':
            pass
        else:
            setattr(self, name, value)

    def exists(self):
        """
        Returns True if the key exists
        
        @rtype: bool
        @return: Whether the key exists on S3
        """
        return bool(self.bucket.lookup(self.name))

    def delete(self):
        """
        Delete this key from S3
        """
        return self.bucket.delete_key(self.name)

    def get_metadata(self, name):
        return self.metadata.get(name)

    def set_metadata(self, name, value):
        self.metadata[name] = value

    def update_metadata(self, d):
        self.metadata.update(d)
    
    # convenience methods for setting/getting ACL
    def set_acl(self, acl_str):
        if self.bucket != None:
            self.bucket.set_acl(acl_str, self.name)

    def get_acl(self):
        if self.bucket != None:
            return self.bucket.get_acl(self.name)

    def get_xml_acl(self):
        if self.bucket != None:
            return self.bucket.get_xml_acl(self.name)

    def set_xml_acl(self, acl_str):
        if self.bucket != None:
            return self.bucket.set_xml_acl(acl_str, self.name)
        
    def make_public(self):
        response = self.bucket.connection.make_request('PUT', self.bucket.name, self.name,
                headers={'x-amz-acl': 'public-read'}, query_args='acl')
        body = response.read()
        if response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)

    def generate_url(self, expires_in, method='GET', headers=None, query_auth=True, force_http=False):
        """
        Generate a URL to access this key.
        
        @type expires_in: int
        @param expires_in: How long the url is valid for
        
        @type method: string
        @param method: The method to use for retrieving the file (default is GET)
        
        @type headers: dict
        @param headers: Any headers to pass along in the request
        
        @type query_auth: bool
        @param query_auth: 
        
        @rtype: string
        @return: The URL to access the key
        """
        return self.bucket.connection.generate_url(expires_in, method,
                                                   self.bucket.name, self.name,
                                                   headers, query_auth, force_http)

    def send_file(self, fp, headers=None, cb=None, num_cb=10):
        """
        Upload a file to a key into a bucket on S3.
        
        @type fp: file
        @param fp: The file pointer to upload
        
        @type headers: dict
        @param headers: The headers to pass along with the PUT request
        
        @type cb: function
        @param cb: a callback function that will be called to report
                    progress on the upload.  The callback should accept two integer
                    parameters, the first representing the number of bytes that have
                    been successfully transmitted to S3 and the second representing
                    the total number of bytes that need to be transmitted.
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        """
        def sender(http_conn, method, path, data, headers):
            http_conn.putrequest(method, path)
            for key in headers:
                http_conn.putheader(key, headers[key])
            http_conn.endheaders()
            fp.seek(0)
            save_debug = self.bucket.connection.debug
            self.bucket.connection.debug = 0
            if cb:
                if num_cb > 2:
                    cb_count = self.size / self.BufferSize / (num_cb-2)
                else:
                    cb_count = 0
                i = total_bytes = 0
                cb(total_bytes, self.size)
            l = fp.read(self.BufferSize)
            while len(l) > 0:
                http_conn.send(l)
                if cb:
                    total_bytes += len(l)
                    i += 1
                    if i == cb_count:
                        cb(total_bytes, self.size)
                        i = 0
                l = fp.read(self.BufferSize)
            if cb:
                cb(total_bytes, self.size)
            response = http_conn.getresponse()
            body = response.read()
            fp.seek(0)
            self.bucket.connection.debug = save_debug
            if response.status == 500 or response.status == 503 or \
                    response.getheader('location'):
                # we'll try again
                return response
            elif response.status >= 200 and response.status <= 299:
                self.etag = response.getheader('etag')
                if self.etag != '"%s"'  % self.md5:
                    raise S3DataError('ETag from S3 did not match computed MD5')
                return response
            else:
                raise S3ResponseError(response.status, response.reason, body)

        if not headers:
            headers = {}
        else:
            headers = headers.copy()
        headers['User-Agent'] = UserAgent
        headers['Content-MD5'] = self.base64md5
        if headers.has_key('Content-Type'):
            self.content_type = headers['Content-Type']
        elif self.path:
            self.content_type = mimetypes.guess_type(self.path)[0]
            if self.content_type == None:
                self.content_type = self.DefaultContentType
            headers['Content-Type'] = self.content_type
        else:
            headers['Content-Type'] = self.content_type
        headers['Content-Length'] = self.size
        headers['Expect'] = '100-Continue'
        headers = boto.utils.merge_meta(headers, self.metadata)
        return self.bucket.connection.make_request('PUT', self.bucket.name,
                self.name, headers, sender=sender)

    def _compute_md5(self, fp):
        """
        @type fp: file
        @param fp: File pointer to the file to MD5 hash
        
        @rtype: string
        @return: MD5 Hash of the file in fp
        """
        m = md5()
        s = fp.read(self.BufferSize)
        while s:
            m.update(s)
            s = fp.read(self.BufferSize)
        self.md5 = m.hexdigest()
        self.base64md5 = base64.encodestring(m.digest())
        if self.base64md5[-1] == '\n':
            self.base64md5 = self.base64md5[0:-1]
        self.size = fp.tell()
        fp.seek(0)

    def set_contents_from_file(self, fp, headers=None, replace=True, cb=None, num_cb=10, policy=None):
        """
        Store an object in S3 using the name of the Key object as the
        key in S3 and the contents of the file pointed to by 'fp' as the
        contents.
        
        @type fp: file
        @param fp: the file whose contents to upload
        
        @type headers: dict
        @param headers: additional HTTP headers that will be sent with the PUT request.

        @type replace: bool
        @param replace: If this parameter is False, the method
                        will first check to see if an object exists in the
                        bucket with the same key.  If it does, it won't
                        overwrite it.  The default value is True which will
                        overwrite the object.
                    
        @type cb: function
        @param cb: a callback function that will be called to report
                    progress on the upload.  The callback should accept two integer
                    parameters, the first representing the number of bytes that have
                    been successfully transmitted to S3 and the second representing
                    the total number of bytes that need to be transmitted.
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.

        @type policy: L{CannedACLString<boto.s3.acl.CannedACLStrings>}
        @param policy: A canned ACL policy that will be applied to the new key in S3.
             
        """
        if policy:
            if headers:
                headers['x-amz-acl'] = policy
            else:
                headers = {'x-amz-acl' : policy}
        if hasattr(fp, 'name'):
            self.path = fp.name
        if self.bucket != None:
            self._compute_md5(fp)
            if self.name == None:
                self.name = self.md5
            if not replace:
                k = self.bucket.lookup(self.name)
                if k:
                    return
            self.send_file(fp, headers, cb, num_cb)

    def set_contents_from_filename(self, filename, headers=None, replace=True, cb=None, num_cb=10, policy=None):
        """
        Store an object in S3 using the name of the Key object as the
        key in S3 and the contents of the file named by 'filename'.
        See set_contents_from_file method for details about the
        parameters.
        
        @type filename: string
        @param filename: The name of the file that you want to put onto S3
        
        @type headers: dict
        @param headers: Additional headers to pass along with the request to AWS.
        
        @type replace: bool
        @param replace: If True, replaces the contents of the file if it already exists.
        
        @type cb: function
        @param cb: (optional) a callback function that will be called to report
             progress on the download.  The callback should accept two integer
             parameters, the first representing the number of bytes that have
             been successfully transmitted from S3 and the second representing
             the total number of bytes that need to be transmitted.        
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        @type policy: L{CannedACLString<boto.s3.acl.CannedACLStrings>}
        @param policy: A canned ACL policy that will be applied to the new key in S3.
             
        """
        fp = open(filename, 'rb')
        self.set_contents_from_file(fp, headers, replace, cb, num_cb, policy)
        fp.close()

    def set_contents_from_string(self, s, headers=None, replace=True, cb=None, num_cb=10, policy=None):
        """
        Store an object in S3 using the name of the Key object as the
        key in S3 and the string 's' as the contents.
        See set_contents_from_file method for details about the
        parameters.
        
        @type headers: dict
        @param headers: Additional headers to pass along with the request to AWS.
        
        @type replace: bool
        @param replace: If True, replaces the contents of the file if it already exists.
        
        @type cb: function
        @param cb: (optional) a callback function that will be called to report
             progress on the download.  The callback should accept two integer
             parameters, the first representing the number of bytes that have
             been successfully transmitted from S3 and the second representing
             the total number of bytes that need to be transmitted.        
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        @type policy: L{CannedACLString<boto.s3.acl.CannedACLStrings>}
        @param policy: A canned ACL policy that will be applied to the new key in S3.
             
        """
        fp = StringIO.StringIO(s)
        self.set_contents_from_file(fp, headers, replace, cb, num_cb, policy)
        fp.close()

    def get_file(self, fp, headers=None, cb=None, num_cb=10, torrent=False):
        """
        Retrieves a file from an S3 Key
        
        @type fp: file
        @param fp: File pointer to put the data into
        
        @type headers: string
        @param: headers to send when retrieving the files
        
        @type cb: function
        @param cb: (optional) a callback function that will be called to report
             progress on the download.  The callback should accept two integer
             parameters, the first representing the number of bytes that have
             been successfully transmitted from S3 and the second representing
             the total number of bytes that need to be transmitted.
        
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        @type torrent: bool
        @param torrent: Flag for whether to get a torrent for the file
        """
        if cb:
            if num_cb > 2:
                cb_count = self.size / self.BufferSize / (num_cb-2)
            else:
                cb_count = 0
            i = total_bytes = 0
            cb(total_bytes, self.size)
        save_debug = self.bucket.connection.debug
        if self.bucket.connection.debug == 1:
            self.bucket.connection.debug = 0
        
        if torrent: torrent = "torrent"
        self.open('r', headers, query_args=torrent)
        for bytes in self:
            fp.write(bytes)
            if cb:
                total_bytes += len(bytes)
                i += 1
                if i == cb_count:
                    cb(total_bytes, self.size)
                    i = 0
        if cb:
            cb(total_bytes, self.size)
        self.close()
        self.bucket.connection.debug = save_debug

    def get_torrent_file(self, fp, headers=None, cb=None, num_cb=10):
        """
        Get a torrent file (see to get_file)
        
        @type fp: file
        @param fp: The file pointer of where to put the torrent
        
        @type headers: dict
        @param headers: Headers to be passed
        
        @type cb: function
        @param cb: Callback function to call on retrieved data
        
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        """
        return self.get_file(fp, headers, cb, num_cb, torrent=True)
    
    def get_contents_to_file(self, fp, headers=None, cb=None, num_cb=10, torrent=False):
        """
        Retrieve an object from S3 using the name of the Key object as the
        key in S3.  Write the contents of the object to the file pointed
        to by 'fp'.
        
        @type fp: File -like object
        @param fp:
        
        @type headers: dict
        @param headers: additional HTTP headers that will be sent with the GET request.
        
        @type cb: function
        @param cb: (optional) a callback function that will be called to report
             progress on the download.  The callback should accept two integer
             parameters, the first representing the number of bytes that have
             been successfully transmitted from S3 and the second representing
             the total number of bytes that need to be transmitted.
             
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        @type torrent: bool
        @param torrent: If True, returns the contents of a torrent file as a string.

        """
        if self.bucket != None:
            self.get_file(fp, headers, cb, num_cb, torrent=torrent)

    def get_contents_to_filename(self, filename, headers=None, cb=None, num_cb=10, torrent=False):
        """
        Retrieve an object from S3 using the name of the Key object as the
        key in S3.  Store contents of the object to a file named by 'filename'.
        See get_contents_to_file method for details about the
        parameters.
        
        @type filename: string
        @param filename: The filename of where to put the file contents
        
        @type headers: dict
        @param headers: Any additional headers to send in the request
        
        @type cb: function
        @param cb: (optional) a callback function that will be called to report
             progress on the download.  The callback should accept two integer
             parameters, the first representing the number of bytes that have
             been successfully transmitted from S3 and the second representing
             the total number of bytes that need to be transmitted.
             
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        @type torrent: bool
        @param torrent: If True, returns the contents of a torrent file as a string.
        
        """
        fp = open(filename, 'wb')
        self.get_contents_to_file(fp, headers, cb, num_cb, torrent=torrent)
        fp.close()
        # if last_modified date was sent from s3, try to set file's timestamp
        if self.last_modified != None:
            try:
                modified_tuple = rfc822.parsedate_tz(self.last_modified)
                modified_stamp = int(rfc822.mktime_tz(modified_tuple))
                os.utime(fp.name, (modified_stamp, modified_stamp))
            except Exception: pass

    def get_contents_as_string(self, headers=None, cb=None, num_cb=10, torrent=False):
        """
        Retrieve an object from S3 using the name of the Key object as the
        key in S3.  Return the contents of the object as a string.
        See get_contents_to_file method for details about the
        parameters.
        
        @type headers: dict
        @param headers: Any additional headers to send in the request
        
        @type cb: function
        @param cb: (optional) a callback function that will be called to report
             progress on the download.  The callback should accept two integer
             parameters, the first representing the number of bytes that have
             been successfully transmitted from S3 and the second representing
             the total number of bytes that need to be transmitted.

        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
                    
        @type cb: int
        @param num_cb: (optional) If a callback is specified with the cb parameter
             this parameter determines the granularity of the callback by defining
             the maximum number of times the callback will be called during the file transfer.  
             
        @type torrent: bool
        @param torrent: If True, returns the contents of a torrent file as a string.
        
        @rtype: string
        @returns: The contents of the file as a string
        """
        fp = StringIO.StringIO()
        self.get_contents_to_file(fp, headers, cb, num_cb, torrent=torrent)
        return fp.getvalue()

    def add_email_grant(self, permission, email_address):
        """
        Convenience method that provides a quick way to add an email grant to a key.
        This method retrieves the current ACL, creates a new grant based on the parameters
        passed in, adds that grant to the ACL and then PUT's the new ACL back to S3.
        
        @type permission: string
        @param permission: The permission being granted.  Should be one of:
                            READ|WRITE|READ_ACP|WRITE_ACP|FULL_CONTROL
                            See http://docs.amazonwebservices.com/AmazonS3/2006-03-01/UsingAuthAccess.html
                            for more details on permissions.
        
        @type email_address: string
        @param email_address: The email address associated with the AWS account your are granting
                                the permission to.
        """
        policy = self.get_acl()
        policy.acl.add_email_grant(permission, email_address)
        self.set_acl(policy)

    def add_user_grant(self, permission, user_id):
        """
        Convenience method that provides a quick way to add a canonical user grant to a key.
        This method retrieves the current ACL, creates a new grant based on the parameters
        passed in, adds that grant to the ACL and then PUT's the new ACL back to S3.
        
        @type permission: string
        @param permission: The permission being granted.  Should be one of:
                            READ|WRITE|READ_ACP|WRITE_ACP|FULL_CONTROL
                            See http://docs.amazonwebservices.com/AmazonS3/2006-03-01/UsingAuthAccess.html
                            for more details on permissions.
        
        @type user_id: string
        @param user_id: The canonical user id associated with the AWS account your are granting
                        the permission to.
        """
        policy = self.get_acl()
        policy.acl.add_user_grant(permission, user_id)
        self.set_acl(policy)



# Copyright (c) 2006-2008 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import uuid

class Distribution:

    def __init__(self, connection=None, config=None, domain_name='',
                 id='', last_modified_time=None, status=''):
        self.connection = connection
        self.config = config
        self.domain_name = domain_name
        self.id = id
        self.last_modified_time = last_modified_time
        self.status = status
        self.etag = None
        
    def startElement(self, name, attrs, connection):
        if name == 'DistributionConfig':
            self.config = DistributionConfig()
            return self.config
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'Id':
            self.id = value
        elif name == 'LastModifiedTime':
            self.last_modified_time = value
        elif name == 'Status':
            self.status = value
        elif name == 'DomainName':
            self.domain_name = value
        else:
            setattr(self, name, value)

    def update(self, enabled=None, cnames=None, comment=None):
        new_config = DistributionConfig(self.connection, self.config.origin,
                                        self.config.enabled, self.config.caller_reference,
                                        self.config.cnames, self.config.comment)
        if enabled != None:
            new_config.enabled = enabled
        if cnames != None:
            new_config.cnames = cnames
        if comment != None:
            new_config.comment = comment
        self.etag = self.connection.set_distribution_config(self.id, self.etag, new_config)
        self.config = new_config

    def enable(self):
        self.update(enabled=True)

    def disable(self):
        self.update(enabled=False)

    def delete(self):
        self.connection.delete_distribution(self.id, self.etag)
            
class DistributionConfig:

    def __init__(self, connection=None, origin='', enabled=False,
                 caller_reference='', cnames=None, comment=''):
        self.connection = connection
        self.origin = origin
        self.enabled = enabled
        if caller_reference:
            self.caller_reference = caller_reference
        else:
            self.caller_reference = str(uuid.uuid4())
        self.cnames = []
        if cnames:
            self.cnames = cnames
        self.comment = comment

    def to_xml(self):
        s = '<?xml version="1.0" encoding="UTF-8"?>\n'
        s += '<DistributionConfig xmlns="http://cloudfront.amazonaws.com/doc/2008-06-30/">\n'
        s += '  <Origin>%s</Origin>\n' % self.origin
        s += '  <CallerReference>%s</CallerReference>\n' % self.caller_reference
        for cname in self.cnames:
            s += '  <CNAME>%s</CNAME>\n' % cname
        if self.comment:
            s += '  <Comment>%s</Comment>\n' % self.comment
        s += '  <Enabled>'
        if self.enabled:
            s += 'true'
        else:
            s += 'false'
        s += '</Enabled>\n'
        s += '</DistributionConfig>\n'
        return s

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'CNAME':
            self.cnames.append(value)
        elif name == 'Origin':
            self.origin = value
        elif name == 'Comment':
            self.comment = value
        elif name == 'Enabled':
            if value.lower() == 'true':
                self.enabled = True
            else:
                self.enabled = False
        elif name == 'CallerReference':
            self.caller_reference = value
        else:
            setattr(self, name, value)

            
class DistributionSummary:

    def __init__(self, connection=None, domain_name='', id='',
                 last_modified_time=None, status='', origin='',
                 cname='', comment='', enabled=False):
        self.connection = connection
        self.domain_name = domain_name
        self.id = id
        self.last_modified_time = last_modified_time
        self.status = status
        self.origin = origin
        self.enabled = enabled
        self.cnames = []
        if cname:
            self.cnames.append(cname)
        self.comment = comment
        self.etag = None

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Id':
            self.id = value
        elif name == 'Status':
            self.status = value
        elif name == 'LastModifiedTime':
            self.last_modified_time = value
        elif name == 'DomainName':
            self.domain_name = value
        elif name == 'Origin':
            self.origin = value
        elif name == 'CNAME':
            self.cnames.append(value)
        elif name == 'Comment':
            self.comment = value
        elif name == 'Enabled':
            if value.lower() == 'true':
                self.enabled = True
            else:
                self.enabled = False
        else:
            setattr(self, name, value)

    def get_distribution(self):
        return self.connection.get_distribution_info(self.id)

        



# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import boto
from boto import handler
from boto.resultset import ResultSet
from boto.s3.acl import Policy, CannedACLStrings, ACL, Grant
from boto.s3.user import User
from boto.s3.key import Key
from boto.s3.prefix import Prefix
from boto.exception import S3ResponseError, S3PermissionsError, S3CopyError
from boto.s3.bucketlistresultset import BucketListResultSet
import boto.utils
import xml.sax
import urllib

S3Permissions = ['READ', 'WRITE', 'READ_ACP', 'WRITE_ACP', 'FULL_CONTROL']

class Bucket:

    BucketLoggingBody = """<?xml version="1.0" encoding="UTF-8"?>
       <BucketLoggingStatus xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
         <LoggingEnabled>
           <TargetBucket>%s</TargetBucket>
           <TargetPrefix>%s</TargetPrefix>
         </LoggingEnabled>
       </BucketLoggingStatus>"""
    
    EmptyBucketLoggingBody = """<?xml version="1.0" encoding="UTF-8"?>
       <BucketLoggingStatus xmlns="http://s3.amazonaws.com/doc/2006-03-01/">
       </BucketLoggingStatus>"""

    LoggingGroup = 'http://acs.amazonaws.com/groups/s3/LogDelivery'

    def __init__(self, connection=None, name=None, key_class=Key):
        self.name = name
        self.connection = connection
        self.key_class = key_class

    def __repr__(self):
        return '<Bucket: %s>' % self.name

    def __iter__(self):
        return iter(BucketListResultSet(self))

    def __contains__(self, key_name):
       return not (self.get_key(key_name) is None)

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Name':
            self.name = value
        elif name == 'CreationDate':
            self.creation_date = value
        else:
            setattr(self, name, value)

    def set_key_class(self, key_class):
        """
        Set the Key class associated with this bucket.  By default, this
        would be the boto.s3.key.Key class but if you want to subclass that
        for some reason this allows you to associate your new class with a
        bucket so that when you call bucket.new_key() or when you get a listing
        of keys in the bucket you will get an instances of your key class
        rather than the default.
        
        @type key_class: class
        @param key_class: A subclass of Key that can be more specific
        """
        self.key_class = key_class

    def lookup(self, key_name):
        """
        Deprecated: Please use get_key method.
        
        @type key_name: string
        @param key_name: The name of the key to retrieve
        
        @rtype: L{Key<boto.s3.key.Key>}
        @returns: A Key object from this bucket.
        """
        return self.get_key(key_name)
        
    def get_key(self, key_name):
        """
        Check to see if a particular key exists within the bucket.  This
        method uses a HEAD request to check for the existance of the key.
        Returns: An instance of a Key object or None
        
        @type key_name: string
        @param key_name: The name of the key to retrieve
        
        @rtype: L{Key<boto.s3.key.Key>}
        @returns: A Key object from this bucket.
        """
        response = self.connection.make_request('HEAD', self.name, key_name)
        if response.status == 200:
            body = response.read()
            k = self.key_class(self)
            k.metadata = boto.utils.get_aws_metadata(response.msg)
            k.etag = response.getheader('etag')
            k.content_type = response.getheader('content-type')
            k.content_encoding = response.getheader('content-encoding')
            k.last_modified = response.getheader('last-modified')
            k.size = int(response.getheader('content-length'))
            k.name = key_name
            return k
        else:
            if response.status == 404:
                body = response.read()
                return None
            else:
                raise S3ResponseError(response.status, response.reason, '')

    def list(self, prefix='', delimiter='', marker=''):
        """
        List key objects within a bucket.  This returns an instance of an
        BucketListResultSet that automatically handles all of the result
        paging, etc. from S3.  You just need to keep iterating until
        there are no more results.
        Called with no arguments, this will return an iterator object across
        all keys within the bucket.
        
        @type prefix: string
        @param prefix: allows you to limit the listing to a particular
                        prefix.  For example, if you call the method with prefix='/foo/'
                        then the iterator will only cycle through the keys that begin with
                        the string '/foo/'.
                        
        @type delimiter: string
        @param delimiter: can be used in conjunction with the prefix
                        to allow you to organize and browse your keys hierarchically. See:
                        http://docs.amazonwebservices.com/AmazonS3/2006-03-01/
                        for more details.
                        
        @type marker: string
        @param marker: The "marker" of where you are in the result set
        
        @rtype: L{BucketListResultSet<boto.s3.bucketlistresultset.BucketListResultSet>}
        @return: an instance of a BucketListResultSet that handles paging, etc
        """
        return BucketListResultSet(self, prefix, delimiter, marker)

    def get_all_keys(self, headers=None, **params):
        """
        A lower-level method for listing contents of a bucket.  This closely models the actual S3
        API and requires you to manually handle the paging of results.  For a higher-level method
        that handles the details of paging for you, you can use the list method.
        
        @type maxkeys: int
        @param maxkeys: The maximum number of keys to retrieve
        
        @type prefix: string
        @param prefix: The prefix of the keys you want to retrieve
        
        @type marker: string
        @param marker: The "marker" of where you are in the result set
        
        @type delimiter: string 
        @param delimiter: "If this optional, Unicode string parameter is included with your request, then keys that contain the same string between the prefix and the first occurrence of the delimiter will be rolled up into a single result element in the CommonPrefixes collection. These rolled-up keys are not returned elsewhere in the response."

        @rtype: ResultSet
        @return: The result from S3 listing the keys requested
        
        """
        l = []
        for k,v in params.items():
            if  k == 'maxkeys':
                k = 'max-keys'
            if isinstance(v, unicode):
                v = v.encode('utf-8')
            if v:
                l.append('%s=%s' % (urllib.quote(k), urllib.quote(str(v))))
        if len(l):
            s = '&'.join(l)
        else:
            s = None
        response = self.connection.make_request('GET', self.name,
                headers=headers, query_args=s)
        body = response.read()
        boto.log.debug(body)
        if response.status == 200:
            rs = ResultSet([('Contents', self.key_class),
                            ('CommonPrefixes', Prefix)])
            h = handler.XmlHandler(rs, self)
            xml.sax.parseString(body, h)
            return rs
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def new_key(self, key_name=None):
        """
        Creates a new key
        
        @type key_name: string
        @param key_name: The name of the key to create
        
        @rtype: L{Key<boto.s3.key.Key>} or subclass
        @returns: An instance of the newly created key object
        """
        return self.key_class(self, key_name)

    def generate_url(self, expires_in, method='GET', headers=None, force_http=False):
        return self.connection.generate_url(expires_in, method, self.name, headers=headers,
                                            force_http=force_http)

    def delete_key(self, key_name):
        """
        Deletes a key from the bucket.
        
        @type key_name: string
        @param key_name: The key name to delete
        """
        response = self.connection.make_request('DELETE', self.name, key_name)
        body = response.read()
        if response.status != 204:
            raise S3ResponseError(response.status, response.reason, body)

    def copy_key(self, new_key_name, src_bucket_name, src_key_name, metadata=None):
        """
        Create a new key in the bucket by copying another existing key.

        @type new_key_name: string
        @param new_key_name: The name of the new key

        @type src_bucket_name: string
        @param src_bucket_name: The name of the source bucket

        @type src_key_name: string
        @param src_key_name: The name of the source key

        @type metadata: dict
        @param metadata: Metadata to be associated with new key.
                         If metadata is supplied, it will replace the
                         metadata of the source key being copied.
                         If no metadata is supplied, the source key's
                         metadata will be copied to the new key.

        @rtype: L{Key<boto.s3.key.Key>} or subclass
        @returns: An instance of the newly created key object
        """
        if metadata:
            headers = {'x-amz-copy-source' : '%s/%s' % (src_bucket_name, src_key_name),
                       'x-amz-metadata-directive' : 'REPLACE'}
            headers = boto.utils.merge_meta(headers, metadata)
        else:
            headers = {'x-amz-copy-source' : '%s/%s' % (src_bucket_name, src_key_name),
                       'x-amz-metadata-directive' : 'COPY'}
        response = self.connection.make_request('PUT', self.name, new_key_name,
                                                headers=headers)
        body = response.read()
        if response.status == 200:
            key = self.new_key(new_key_name)
            h = handler.XmlHandler(key, self)
            xml.sax.parseString(body, h)
            if hasattr(key, 'Error'):
                raise S3CopyError(key.Code, key.Message, body)
            return key
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_canned_acl(self, acl_str, key_name=''):
        assert acl_str in CannedACLStrings
        response = self.connection.make_request('PUT', self.name, key_name,
                headers={'x-amz-acl': acl_str}, query_args='acl')
        body = response.read()
        if response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)

    def get_xml_acl(self, key_name=''):
        response = self.connection.make_request('GET', self.name, key_name,
                                                query_args='acl')
        body = response.read()
        if response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)
        return body

    def set_xml_acl(self, acl_str, key_name=''):
        response = self.connection.make_request('PUT', self.name, key_name,
                data=acl_str, query_args='acl')
        body = response.read()
        if response.status != 200:
            raise S3ResponseError(response.status, response.reason, body)

    def set_acl(self, acl_or_str, key_name=''):
        if isinstance(acl_or_str, Policy):
            self.set_xml_acl(acl_or_str.to_xml(), key_name)
        else:
            self.set_canned_acl(acl_or_str, key_name)

    def get_acl(self, key_name=''):
        response = self.connection.make_request('GET', self.name, key_name,
                query_args='acl')
        body = response.read()
        if response.status == 200:
            policy = Policy(self)
            h = handler.XmlHandler(policy, self)
            xml.sax.parseString(body, h)
            return policy
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def make_public(self, recursive=False):
        self.set_canned_acl('public-read')
        if recursive:
            for key in self:
                self.set_canned_acl('public-read', key.name)

    def add_email_grant(self, permission, email_address, recursive=False):
        """
        Convenience method that provides a quick way to add an email grant to a bucket.
        This method retrieves the current ACL, creates a new grant based on the parameters
        passed in, adds that grant to the ACL and then PUT's the new ACL back to S3.
        Inputs:
            permission - The permission being granted.  Should be one of:
                         READ|WRITE|READ_ACP|WRITE_ACP|FULL_CONTROL
                         See http://docs.amazonwebservices.com/AmazonS3/2006-03-01/UsingAuthAccess.html
                         for more details on permissions.
            email_address - The email address associated with the AWS account your are granting
                            the permission to.
            recursive - A boolean value to controls whether the command will apply the
                        grant to all keys within the bucket or not.  The default value is False.
                        By passing a True value, the call will iterate through all keys in the
                        bucket and apply the same grant to each key.
                        CAUTION: If you have a lot of keys, this could take a long time!
        Returns:
            Nothing
        """
        if permission not in S3Permissions:
            raise S3PermissionsError('Unknown Permission: %s' % permission)
        policy = self.get_acl()
        policy.acl.add_email_grant(permission, email_address)
        self.set_acl(policy)
        if recursive:
            for key in self:
                key.add_email_grant(permission, email_address)

    def add_user_grant(self, permission, user_id, recursive=False):
        """
        Convenience method that provides a quick way to add a canonical user grant to a bucket.
        This method retrieves the current ACL, creates a new grant based on the parameters
        passed in, adds that grant to the ACL and then PUT's the new ACL back to S3.
        
        @type permission: string
        @param permission:  The permission being granted.  Should be one of:
                            READ|WRITE|READ_ACP|WRITE_ACP|FULL_CONTROL
                            See http://docs.amazonwebservices.com/AmazonS3/2006-03-01/UsingAuthAccess.html
                            for more details on permissions.
                            
        @type user_id: string
        @param user_id:     The canonical user id associated with the AWS account your are granting
                            the permission to.
                            
        @type recursive: bool
        @param recursive:   A boolean value that controls whether the command will apply the
                            grant to all keys within the bucket or not.  The default value is False.
                            By passing a True value, the call will iterate through all keys in the
                            bucket and apply the same grant to each key.
                            CAUTION: If you have a lot of keys, this could take a long time!
        """
        if permission not in S3Permissions:
            raise S3PermissionsError('Unknown Permission: %s' % permission)
        policy = self.get_acl()
        policy.acl.add_user_grant(permission, user_id)
        self.set_acl(policy)
        if recursive:
            for key in self:
                key.add_user_grant(permission, user_id)

    def list_grants(self):
        policy = self.get_acl()
        return policy.acl.grants

#    def get_location(self):
#        location = _XmlInterpreter(
#                string_attrs={'LocationConstraint': 'location'})
#        body = self.connection.checked_request(location, self,
#                'GET', self.name, query_args='location').location
#        return parse_xml_string(body, location, self)

    def enable_logging(self, target_bucket, target_prefix=''):
        if isinstance(target_bucket, Bucket):
            target_bucket = target_bucket.name
        body = self.BucketLoggingBody % (target_bucket, target_prefix)
        response = self.connection.make_request('PUT', self.name, data=body,
                query_args='logging')
        body = response.read()
        if response.status == 200:
            return True
        else:
            raise S3ResponseError(response.status, response.reason, body)
        
    def disable_logging(self):
        body = self.EmptyBucketLoggingBody
        response = self.connection.make_request('PUT', self.name, data=body,
                query_args='logging')
        body = response.read()
        if response.status == 200:
            return True
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def get_logging_status(self):
        response = self.connection.make_request('GET', self.name,
                query_args='logging')
        body = response.read()
        if response.status == 200:
            return body
        else:
            raise S3ResponseError(response.status, response.reason, body)

    def set_as_logging_target(self):
        policy = self.get_acl()
        g1 = Grant(permission='WRITE', type='Group', uri=self.LoggingGroup)
        g2 = Grant(permission='READ_ACP', type='Group', uri=self.LoggingGroup)
        policy.acl.add_grant(g1)
        policy.acl.add_grant(g2)
        self.set_acl(policy)

    def delete(self):
        return self.connection.delete_bucket(self.name)



import SocketServer, os, datetime, sys, random, time
import simplejson

class LQSCommand:

    def __init__(self, line):
        self.raw_line = line
        self.line = self.raw_line.strip()
        l = self.line.split(' ')
        self.name = l[0]
        if len(l) > 1:
            self.args = [arg for arg in l[1:] if arg]
        else:
            self.args = []

class LQSMessage(dict):

    def __init__(self, item=None, args=None, jsonvalue=None):
        dict.__init__(self)
        if jsonvalue:
            self.decode(jsonvalue)
        else:
            self['id'] = '%d_%d' % (int(time.time()), int(random.random()*1000000))
            self['item'] = item
            self['args'] = args

    def encode(self):
        return simplejson.dumps(self)

    def decode(self, value):
        self.update(simplejson.loads(value))

    def is_empty(self):
        if self['item'] == None:
            return True
        return False

class LQSServer(SocketServer.UDPServer):

    PORT = 5151
    TIMEOUT = 30
    MAXSIZE = 8192

    def __init__(self, server_address, RequestHandlerClass, iterator, args=None):
        server_address = (server_address, self.PORT)
        SocketServer.UDPServer.__init__(self, server_address, RequestHandlerClass)
        self.count = 0
        self.iterator = iterator
        self.args = args
        self.start = datetime.datetime.now()
        self.end = None
        self.extant = []

class LQSHandler(SocketServer.DatagramRequestHandler):

    def get_cmd(self):
        return LQSCommand(self.rfile.readline())

    def build_msg(self):
        if not self.server.iterator:
            return LQSMessage(None)
        try:
            item = self.server.iterator.next()
            msg = LQSMessage(item, self.server.args)
            return msg
        except StopIteration:
            self.server.iterator = None
            return LQSMessage(None)

    def respond(self, msg):
        self.wfile.write(msg.encode())

    def check_extant(self):
        if len(self.server.extant) == 0 and not self.server.iterator:
            self.server.end = datetime.datetime.now()
            delta = self.server.end - self.server.start
            print 'Total Processing Time: %s' % delta
            print 'Total Messages Processed: %d' % self.server.count

    def do_debug(self, cmd):
        args = {'extant' : self.server.extant,
                'count' : self.server.count}
        msg = LQSMessage('debug', args)
        self.respond(msg)

    def do_next(self, cmd):
        out_msg = self.build_msg()
        if not out_msg.is_empty():
            self.server.count += 1
            self.server.extant.append(out_msg['id'])
        self.respond(out_msg)

    def do_delete(self, cmd):
        if len(cmd.args) != 1:
            self.error(cmd, 'delete command requires message id')
        else:
            mid = cmd.args[0]
            try:
                self.server.extant.remove(mid)
            except ValueError:
                self.error(cmd, 'message id not found')
            args = {'deleted' : True}
            msg = LQSMessage(mid, args)
            self.respond(msg)
            self.check_extant()

    def error(self, cmd, error_msg=None):
        args = {'error_msg' : error_msg,
                'cmd_name' : cmd.name,
                'cmd_args' : cmd.args}
        msg = LQSMessage('error', args)
        self.respond(msg)

    def do_stop(self, cmd):
        sys.exit(0)

    def handle(self):
        cmd = self.get_cmd()
        if hasattr(self, 'do_%s' % cmd.name):
            method = getattr(self, 'do_%s' % cmd.name)
            method(cmd)
        else:
            self.error(cmd, 'unrecognized command')

class PersistHandler(LQSHandler):

    def build_msg(self):
        if not self.server.iterator:
            return LQSMessage(None)
        try:
            obj = self.server.iterator.next()
            msg = LQSMessage(obj.id, self.server.args)
            return msg
        except StopIteration:
            self.server.iterator = None
            return LQSMessage(None)

def test_file(path, args=None):
    l = os.listdir(path)
    if not args:
        args = {}
    args['path'] = path
    s = LQSServer('', LQSHandler, iter(l), args)
    print "Awaiting UDP messages on port %d" % s.PORT
    s.serve_forever()

def test_simple(n):
    l = range(0, n)
    s = LQSServer('', LQSHandler, iter(l), None)
    print "Awaiting UDP messages on port %d" % s.PORT
    s.serve_forever()




# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
import StringIO, os
import ConfigParser
import boto

BotoConfigLocations = ['/etc/boto.cfg', os.path.expanduser('~/.boto')]
BotoConfigPath = BotoConfigLocations[0]
UserConfigPath = BotoConfigLocations[1]

class Config(ConfigParser.SafeConfigParser):

    def __init__(self, path=None, fp=None, do_load=True):
        ConfigParser.SafeConfigParser.__init__(self, {'working_dir' : '/mnt/pyami',
                                                      'debug' : '0'})
        if do_load:
            if path:
                self.read(path)
            elif fp:
                self.readfp(fp)
            else:
                self.read(BotoConfigLocations)

    def save_option(self, path, section, option, value):
        """
        Write the specified Section.Option to the config file specified by path.
        Replace any previous value.  If the path doesn't exist, create it.
        Also add the option the the in-memory config.
        """
        config = ConfigParser.SafeConfigParser()
        config.read(path)
        if not config.has_section(section):
            config.add_section(section)
        config.set(section, option, value)
        fp = open(path, 'w')
        config.write(fp)
        fp.close()
        if not self.has_section(section):
            self.add_section(section)
        self.set(section, option, value)

    def save_user_option(self, section, option, value):
        self.save_option(UserConfigPath, section, option, value)

    def save_system_option(self, section, option, value):
        self.save_option(BotoConfigPath, section, option, value)

    def get_instance(self, name, default=None):
        try:
            val = self.get('Instance', name)
        except:
            val = default
        return val

    def get_user(self, name, default=None):
        try:
            val = self.get('User', name)
        except:
            val = default
        return val

    def getint_user(self, name, default=0):
        try:
            val = self.getint('User', name)
        except:
            val = default
        return val

    def get_value(self, section, name, default=None):
        return self.get(section, name, default)

    def get(self, section, name, default=None):
        try:
            val = ConfigParser.SafeConfigParser.get(self, section, name)
        except:
            val = default
        return val
    
    def getint(self, section, name, default=0):
        try:
            val = ConfigParser.SafeConfigParser.getint(self, section, name)
        except:
            val = int(default)
        return val
    
    def getfloat(self, section, name, default=0.0):
        try:
            val = ConfigParser.SafeConfigParser.getfloat(self, section, name)
        except:
            val = float(default)
        return val

    def getbool(self, section, name, default=False):
        if self.has_option(section, name):
            val = self.get(section, name)
            if val.lower() == 'true':
                val = True
            else:
                val = False
        else:
            val = default
        return val
    
    def setbool(self, section, name, value):
        if value:
            self.set(section, name, 'true')
        else:
            self.set(section, name, 'false')
    
    def dump(self):
        s = StringIO.StringIO()
        self.write(s)
        print s.getvalue()

    def dump_safe(self, fp=None):
        if not fp:
            fp = StringIO.StringIO()
        for section in self.sections():
            fp.write('[%s]\n' % section)
            for option in self.options(section):
                if option == 'aws_secret_access_key':
                    fp.write('%s = xxxxxxxxxxxxxxxxxx\n' % option)
                else:
                    fp.write('%s = %s\n' % (option, self.get(section, option)))
    
    def dump_to_sdb(self, domain_name, item_name):
        import simplejson
        sdb = boto.connect_sdb()
        domain = sdb.lookup(domain_name)
        if not domain:
            domain = sdb.create_domain(domain_name)
        item = domain.new_item(item_name)
        item.active = False
        for section in self.sections():
            d = {}
            for option in self.options(section):
                d[option] = self.get(section, option)
            item[section] = simplejson.dumps(d)
        item.save()

    def load_from_sdb(self, domain_name, item_name):
        import simplejson
        sdb = boto.connect_sdb()
        domain = sdb.lookup(domain_name)
        item = domain.get_item(item_name)
        for section in item.keys():
            if not self.has_section(section):
                self.add_section(section)
            d = simplejson.loads(item[section])
            for attr_name in d.keys():
                attr_value = d[attr_name]
                if attr_value == None:
                    attr_value = 'None'
                if isinstance(attr_value, bool):
                    self.setbool(section, attr_name, attr_value)
                else:
                    self.set(section, attr_name, attr_value)



import uuid
import datetime
from boto.mturk.connection import MTurkConnection
from boto.mturk.question import ExternalQuestion

def test():
    q = ExternalQuestion(external_url="http://websort.net/s/F3481C", frame_height=800)
    conn = MTurkConnection(host='mechanicalturk.sandbox.amazonaws.com')
    keywords=['boto', 'test', 'doctest']
    create_hit_rs = conn.create_hit(question=q, lifetime=60*65,max_assignments=2,title="Boto External Question Test", keywords=keywords,reward = 0.05, duration=60*6,approval_delay=60*60, annotation='An annotation from boto external question test', response_groups=['Minimal','HITDetail','HITQuestion','HITAssignmentSummary',])
    assert(create_hit_rs.status == True)

if __name__ == "__main__":
    test()



# Copyright (c) 2006-2008 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Represents a connection to the EC2 service.
"""

import urllib
import xml.sax
import base64
import boto
from boto import config
from boto.connection import AWSQueryConnection
from boto.resultset import ResultSet
from boto.ec2.image import Image, ImageAttribute
from boto.ec2.instance import Reservation, Instance, ConsoleOutput
from boto.ec2.keypair import KeyPair
from boto.ec2.address import Address
from boto.ec2.volume import Volume
from boto.ec2.snapshot import Snapshot
from boto.ec2.zone import Zone
from boto.ec2.securitygroup import SecurityGroup
from boto.ec2.regioninfo import RegionInfo
from boto.exception import EC2ResponseError

class EC2Connection(AWSQueryConnection):

    APIVersion = boto.config.get('Boto', 'ec2_version', '2008-12-01')
    DefaultRegionName = boto.config.get('Boto', 'ec2_region_name', 'us-east-1')
    DefaultRegionEndpoint = boto.config.get('Boto', 'ec2_region_endpoint',
                                            'us-east-1.ec2.amazonaws.com')
    SignatureVersion = '2'
    ResponseError = EC2ResponseError

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                 is_secure=True, port=None, proxy=None, proxy_port=None,
                 proxy_user=None, proxy_pass=None, debug=0,
                 https_connection_factory=None, region=None):
        """
        Init method to create a new connection to EC2.
        
        B{Note:} The host argument is overridden by the host specified in the boto configuration file.        
        """
        if not region:
            region = RegionInfo(self, self.DefaultRegionName, self.DefaultRegionEndpoint)
        self.region = region
        AWSQueryConnection.__init__(self, aws_access_key_id, aws_secret_access_key,
                                    is_secure, port, proxy, proxy_port, proxy_user, proxy_pass,
                                    self.region.endpoint, debug, https_connection_factory)

    def get_params(self):
        """
        Returns a dictionary containing the value of of all of the keyword
        arguments passed when constructing this connection.
        """
        param_names = ['aws_access_key_id', 'aws_secret_access_key', 'is_secure',
                       'port', 'proxy', 'proxy_port', 'proxy_user', 'proxy_pass',
                       'debug', 'https_connection_factory']
        params = {}
        for name in param_names:
            params[name] = getattr(self, name)
        return params

    # Image methods
        
    def get_all_images(self, image_ids=None, owners=None, executable_by=None):
        """
        Retrieve all the EC2 images available on your account.
        
        @type image_ids: list
        @param image_ids: A list of strings with the image IDs wanted
        
        @type owners: list
        @param owners: A list of owner IDs
        
        @type executable_by: 
        @param executable_by: 
        
        @rtype: list
        @return: A list of L{Images<boto.ec2.image.Image>}
        """
        params = {}
        if image_ids:
            self.build_list_params(params, image_ids, 'ImageId')
        if owners:
            self.build_list_params(params, owners, 'Owner')
        if executable_by:
            self.build_list_params(params, executable_by, 'ExecutableBy')
        return self.get_list('DescribeImages', params, [('item', Image)])

    def get_all_kernels(self, kernel_ids=None, owners=None):
        """
        Retrieve all the EC2 kernels available on your account.  Simply filters the list returned
        by get_all_images because EC2 does not provide a way to filter server-side.
        
        @type kernel_ids: list
        @param kernel_ids: A list of strings with the image IDs wanted
        
        @type owners: list
        @param owners: A list of owner IDs
        
        @rtype: list
        @return: A list of L{Images<boto.ec2.image.Image>}
        """
        rs = self.get_all_images(kernel_ids, owners)
        kernels = []
        for image in rs:
            if image.type == 'kernel':
                kernels.append(image)
        return kernels

    def get_all_ramdisks(self, ramdisk_ids=None, owners=None):
        """
        Retrieve all the EC2 ramdisks available on your account.  Simply filters the list returned
        by get_all_images because EC2 does not provide a way to filter server-side.
        
        @type ramdisk_ids: list
        @param ramdisk_ids: A list of strings with the image IDs wanted
        
        @type owners: list
        @param owners: A list of owner IDs
        
        @rtype: list
        @return: A list of L{Images<boto.ec2.image.Image>}
        """
        rs = self.get_all_images(ramdisk_ids, owners)
        ramdisks = []
        for image in rs:
            if image.type == 'ramdisk':
                ramdisks.append(image)
        return ramdisks

    def get_image(self, image_id):
        """
        Shortcut method to retrieve a specific image (AMI).
        
        @type image_id: string
        @param image_id: the ID of the Image to retrieve
        
        @rtype: L{Image<boto.ec2.image.Image>}
        @return: The EC2 Image specified or None if the image is not found
        """
        try:
            return self.get_all_images(image_ids=[image_id])[0]
        except IndexError: # None of those images available
            return None

    def register_image(self, image_location):
        """
        Register an image.

        @type image_location: string
        @param image_location: Full path to your AMI manifest in Amazon S3 storage.

        @rtype: string
        @return: The new image id
        """
        params = {'ImageLocation':image_location}
        rs = self.get_object('RegisterImage', params, ResultSet)
        return rs.imageId
        
    def deregister_image(self, image_id):
        """
        Unregister an AMI.
        
        @type image_id: string
        @param image_id: the ID of the Image to unregister
        
        @rtype: bool
        @return: True if successful
        """
        return self.get_status('DeregisterImage', {'ImageId':image_id})
        
    # ImageAttribute methods
        
    def get_image_attribute(self, image_id, attribute='launchPermission'):
        """
        Gets an attribute from an image.
        See http://docs.amazonwebservices.com/AWSEC2/2008-02-01/DeveloperGuide/ApiReference-Query-DescribeImageAttribute.html
        
        @type image_id: string
        @param image_id: The Amazon image id for which you want info about
        
        @type attribute: string
        @param attribute: The attribute you need information about
        
        @rtype: L{ImageAttribute<boto.ec2.image.ImageAttribute>}
        @return: An ImageAttribute object representing the value of the attribute requested
        """
        params = {'ImageId' : image_id,
                  'Attribute' : attribute}
        return self.get_object('DescribeImageAttribute', params, ImageAttribute)
        
    def modify_image_attribute(self, image_id, attribute='launchPermission',
                               operation='add', user_ids=None, groups=None):
        """
        Changes an attribute of an image.
        See http://docs.amazonwebservices.com/AWSEC2/2008-02-01/DeveloperGuide/ApiReference-Query-ModifyImageAttribute.html
        
        @type image_id: string
        @param image_id: The image id you wish to change
        
        @type attribute: string
        @param attribute: The attribute you wish to change
        
        @type operation: string
        @param operation: Either add or remove (this is required for changing launchPermissions
        
        @type user_ids: list
        @param user_ids: The Amazon IDs of users to add/remove attributes
        
        @type groups: list
        @param groups: The groups to add/remove attributes
        """
        params = {'ImageId' : image_id,
                  'Attribute' : attribute,
                  'OperationType' : operation}
        if user_ids:
            self.build_list_params(params, user_ids, 'UserId')
        if groups:
            self.build_list_params(params, groups, 'UserGroup')
        return self.get_status('ModifyImageAttribute', params)

    def reset_image_attribute(self, image_id, attribute='launchPermission'):
        """
        Resets an attribute of an AMI to its default value.
        See http://docs.amazonwebservices.com/AWSEC2/2008-02-01/DeveloperGuide/ApiReference-Query-ResetImageAttribute.html
        
        @type image_id: string
        @param image_id: ID of the AMI for which an attribute will be described
        
        @type attribute: string
        @param attribute: The attribute to reset
        
        @rtype: bool
        @return: Whether the operation succeeded or not
        """
        params = {'ImageId' : image_id,
                  'Attribute' : attribute}
        return self.get_status('ResetImageAttribute', params)
        
    # Instance methods
        
    def get_all_instances(self, instance_ids=None):
        """
        Retrieve all the instances associated with your account.
        
        @type instance_ids: list
        @param instance_ids: A list of strings of instance IDs
        
        @rtype: list
        @return: A list of L{Instances<boto.ec2.instance.Instance>}
        """
        params = {}
        if instance_ids:
            self.build_list_params(params, instance_ids, 'InstanceId')
        return self.get_list('DescribeInstances', params, [('item', Reservation)])

    def run_instances(self, image_id, min_count=1, max_count=1,
                      key_name=None, security_groups=None,
                      user_data=None, addressing_type=None,
                      instance_type='m1.small', placement=None,
                      kernel_id=None, ramdisk_id=None):
        """
        Runs an image on EC2.
        
        @type image_id: string
        @param image_id: The ID of the image to run
        
        @type min_count: int
        @param min_count: The minimum number of instances to launch
        
        @type max_count: int
        @param max_count: The maximum number of instances to launch
        
        @type key_name: string
        @param key_name: The name of the key pair with which to launch instances
        
        @type security_groups: list of strings
        @param security_groups: The names of the security groups with which to associate instances
        
        @type user_data: string
        @param user_data: The user data passed to the launched instances
        
        @type instance_type: string
        @param instance_type: The type of instance to run (m1.small, m1.large, m1.xlarge)
        
        @type placement: string
        @param placement: The availability zone in which to launch the instances
        
        @type kernel_id: string
        @param kernel_id: The ID of the kernel with which to launch the instances
        
        @type ramdisk_id: string
        @param ramdisk_id: The ID of the RAM disk with which to launch the instances
        
        @rtype: Reservation
        @return: The L{Reservation<boto.ec2.instance.Reservation>} associated with the request for machines
        """
        params = {'ImageId':image_id,
                  'MinCount':min_count,
                  'MaxCount': max_count}
        if key_name:
            params['KeyName'] = key_name
        if security_groups:
            l = []
            for group in security_groups:
                if isinstance(group, SecurityGroup):
                    l.append(group.name)
                else:
                    l.append(group)
            self.build_list_params(params, l, 'SecurityGroup')
        if user_data:
            params['UserData'] = base64.b64encode(user_data)
        if addressing_type:
            params['AddressingType'] = addressing_type
        if instance_type:
            params['InstanceType'] = instance_type
        if placement:
            params['Placement.AvailabilityZone'] = placement
        if kernel_id:
            params['KernelId'] = kernel_id
        if ramdisk_id:
            params['RamdiskId'] = ramdisk_id
        return self.get_object('RunInstances', params, Reservation)
        
    def terminate_instances(self, instance_ids=None):
        """
        Terminate the instances specified
        
        @type instance_ids: list
        @param instance_ids: A list of strings of the Instance IDs to terminate
        
        @rtype: list
        @return: A list of the instances terminated
        """
        params = {}
        if instance_ids:
            self.build_list_params(params, instance_ids, 'InstanceId')
        return self.get_list('TerminateInstances', params, [('item', Instance)])

    def get_console_output(self, instance_id):
        """
        Retrieves the console output for the specified instance.
        See http://docs.amazonwebservices.com/AWSEC2/2008-02-01/DeveloperGuide/ApiReference-Query-GetConsoleOutput.html
        
        @type instance_id: string
        @param instance_id: The instance ID of a running instance on the cloud.
        
        @rtype: L{boto.ec2.instance.ConsoleOutput}
        @return: The console output as a ConsoleOutput object
        """
        params = {}
        self.build_list_params(params, [instance_id], 'InstanceId')
        return self.get_object('GetConsoleOutput', params, ConsoleOutput)

    def reboot_instances(self, instance_ids=None):
        """
        Reboot the specified instances.
        
        @type instance_ids: list
        @param instance_ids: The instances to terminate and reboot
        """
        params = {}
        if instance_ids:
            self.build_list_params(params, instance_ids, 'InstanceId')
        return self.get_status('RebootInstances', params)

    def confirm_product_instance(self, product_code, instance_id):
        params = {'ProductCode' : product_code,
                  'InstanceId' : instance_id}
        rs = self.get_object('ConfirmProductInstance', params, ResultSet)
        return (rs.status, rs.ownerId)

    # Zone methods

    def get_all_zones(self, zones=None):
        """
        Get all Availability Zones associated with the current region.

        @type zones: list
        @param zones: Optional list of zones.  If this list is present,
                      only the Zones associated with these zone names
                      will be returned.

        @rtype: list of L{boto.ec2.zone.Zone}
        @return: The requested Zone objects
        """
        params = {}
        if zones:
            self.build_list_params(params, zones, 'ZoneName')
        return self.get_list('DescribeAvailabilityZones', params, [('item', Zone)])

    # Address methods

    def get_all_addresses(self, addresses=None):
        """
        Get all EIP's associated with the current credentials.

        @type addresses: list
        @param addresses: Optional list of addresses.  If this list is present,
                           only the Addresses associated with these addresses
                           will be returned.

        @rtype: list of L{boto.ec2.address.Address}
        @return: The requested Address objects
        """
        params = {}
        if addresses:
            self.build_list_params(params, addresses, 'PublicIp')
        return self.get_list('DescribeAddresses', params, [('item', Address)])

    def allocate_address(self):
        """
        Allocate a new Elastic IP address and associate it with your account.

        @rtype: L{boto.ec2.address.Address}
        @return: The newly allocated Address
        """
        return self.get_object('AllocateAddress', None, Address)

    def associate_address(self, instance_id, public_ip):
        """
        Associate an Elastic IP address with a currently running instance.

        @type instance_id: string
        @param instance_id: The ID of the instance

        @type public_ip: string
        @param public_ip: The public IP address

        @rtype: bool
        @return: True if successful
        """
        params = {'InstanceId' : instance_id, 'PublicIp' : public_ip}
        return self.get_status('AssociateAddress', params)

    def disassociate_address(self, public_ip):
        """
        Disassociate an Elastic IP address from a currently running instance.

        @type public_ip: string
        @param public_ip: The public IP address

        @rtype: bool
        @return: True if successful
        """
        params = {'PublicIp' : public_ip}
        return self.get_status('DisassociateAddress', params)

    def release_address(self, public_ip):
        """
        Free up an Elastic IP address

        @type public_ip: string
        @param public_ip: The public IP address

        @rtype: bool
        @return: True if successful
        """
        params = {'PublicIp' : public_ip}
        return self.get_status('ReleaseAddress', params)

    # Volume methods

    def get_all_volumes(self, volume_ids=None):
        """
        Get all Volumes associated with the current credentials.

        @type volume_ids: list
        @param volume_ids: Optional list of volume ids.  If this list is present,
                           only the volumes associated with these volume ids
                           will be returned.

        @rtype: list of L{boto.ec2.volume.Volume}
        @return: The requested Volume objects
        """
        params = {}
        if volume_ids:
            self.build_list_params(params, volume_ids, 'VolumeId')
        return self.get_list('DescribeVolumes', params, [('item', Volume)])
        
    def create_volume(self, size, zone, snapshot=None):
        """
        Create a new EBS Volume.

        @type size: int
        @param size: The size of the new volume, in GiB

        @type zone: string or L{boto.ec2.zone.Zone}
        @param zone: The availability zone in which the Volume will be created.

        @type snapshot: string or L{boto.ec2.snapshot.Snapshot}
        @param snapshot: The snapshot from which the new Volume will be created.
        """
        if isinstance(zone, Zone):
            zone = zone.name
        params = {'Size': size, 'AvailabilityZone' : zone}
        if snapshot:
            if isinstance(snapshot, Snapshot):
                snapshot = snapshot.id
            params['SnapshotId'] = snapshot
        return self.get_object('CreateVolume', params, Volume)
        
    def delete_volume(self, volume_id):
        params = {'VolumeId': volume_id}
        return self.get_status('DeleteVolume', params)

    def attach_volume(self, volume_id, instance_id, device):
        params = {'InstanceId' : instance_id,
                  'VolumeId' : volume_id,
                  'Device' : device}
        return self.get_status('AttachVolume', params)

    def detach_volume(self, volume_id, instance_id, device='', force=False):
        params = {'InstanceId' : instance_id,
                  'VolumeId' : volume_id,
                  'Device' : device}
        if force:
            params['Force'] = 'true'
        return self.get_status('DetachVolume', params)

    # Snapshot methods

    def get_all_snapshots(self, snapshot_ids=None):
        """
        Get all EBS Snapshots associated with the current credentials.

        @type snapshot_ids: list
        @param snapshot_ids: Optional list of snapshot ids.  If this list is present,
                           only the Snapshots associated with these snapshot ids
                           will be returned.

        @rtype: list of L{boto.ec2.snapshot.Snapshot}
        @return: The requested Snapshot objects
        """
        params = {}
        if snapshot_ids:
            self.build_list_params(params, snapshot_ids, 'SnapshotId')
        return self.get_list('DescribeSnapshots', params, [('item', Snapshot)])
        
    def create_snapshot(self, volume_id):
        params = {'VolumeId' : volume_id}
        return self.get_object('CreateSnapshot', params, Snapshot)
        
    def delete_snapshot(self, snapshot_id):
        params = {'SnapshotId': snapshot_id}
        return self.get_status('DeleteSnapshot', params)

    # Keypair methods
        
    def get_all_key_pairs(self, keynames=None):
        """
        Get all key pairs associated with your account.
        
        @type keynames: list
        @param keynames: A list of the names of keypairs to retrieve.
                         If not provided, all key pairs will be returned.
        
        @rtype: list
        @return: A list of L{KeyPairs<boto.ec2.keypair.KeyPair>}
        """
        params = {}
        if keynames:
            self.build_list_params(params, keynames, 'KeyName')
        return self.get_list('DescribeKeyPairs', params, [('item', KeyPair)])
    
    def get_key_pair(self, keyname):
        """
        Convenience method to retrieve a specific keypair (KeyPair).
        
        @type image_id: string
        @param image_id: the ID of the Image to retrieve
        
        @rtype: L{KeyPair<boto.ec2.keypair.KeyPair>}
        @return: The KeyPair specified or None if it is not found
        """
        try:
            return self.get_all_key_pairs(keynames=[keyname])[0]
        except IndexError: # None of those key pairs available
            return None
        
    def create_key_pair(self, key_name):
        """
        Create a new key pair for your account.
        This will create the key pair within the region you
        are currently connected to.
        
        @type key_name: string
        @param key_name: The name of the new keypair
        
        @rtype: L{KeyPair<boto.ec2.keypair.KeyPair>}
        @return: The newly created L{KeyPair<boto.ec2.keypair.KeyPair>}.
                 The material attribute of the new KeyPair object
                 will contain the the unencrypted PEM encoded RSA private key.
        """
        params = {'KeyName':key_name}
        return self.get_object('CreateKeyPair', params, KeyPair)
        
    def delete_key_pair(self, key_name):
        """
        Delete a key pair from your account.
        
        @type key_name: string
        @param key_name: The name of the keypair to delete
        """
        params = {'KeyName':key_name}
        return self.get_status('DeleteKeyPair', params)

    # SecurityGroup methods
        
    def get_all_security_groups(self, groupnames=None):
        """
        Get all security groups associated with your account in a region.
        
        @type groupnames: list
        @param groupnames: A list of the names of security groups to retrieve.
                           If not provided, all security groups will be returned.
        
        @rtype: list
        @return: A list of L{SecurityGroups<boto.ec2.securitygroup.SecurityGroup>}
        """
        params = {}
        if groupnames:
            self.build_list_params(params, groupnames, 'GroupName')
        return self.get_list('DescribeSecurityGroups', params, [('item', SecurityGroup)])

    def create_security_group(self, name, description):
        """
        Create a new security group for your account.
        This will create the security group within the region you
        are currently connected to.
        
        @type name: string
        @param name: The name of the new security group
        
        @type description: string
        @param description: The description of the new security group
        
        @rtype: L{SecurityGroup<boto.ec2.securitygroup.SecurityGroup>}
        @return: The newly created L{KeyPair<boto.ec2.keypair.KeyPair>}.
        """
        params = {'GroupName':name, 'GroupDescription':description}
        group = self.get_object('CreateSecurityGroup', params, SecurityGroup)
        group.name = name
        group.description = description
        return group

    def delete_security_group(self, name):
        """
        Delete a security group from your account.
        
        @type key_name: string
        @param key_name: The name of the keypair to delete
        """
        params = {'GroupName':name}
        return self.get_status('DeleteSecurityGroup', params)

    def authorize_security_group(self, group_name, src_security_group_name=None,
                                 src_security_group_owner_id=None,
                                 ip_protocol=None, from_port=None, to_port=None,
                                 cidr_ip=None):
        """
        Add a new rule to an existing security group.
        You need to pass in either src_security_group_name and
        src_security_group_owner_id OR ip_protocol, from_port, to_port,
        and cidr_ip.  In other words, either you are authorizing another
        group or you are authorizing some ip-based rule.
        
        @type group_name: string
        @param group_name: The name of the security group you are adding
                           the rule to.
                           
        @type src_security_group_name: string
        @param src_security_group_name: The name of the security group you are 
                                        granting access to.
                                        
        @type src_security_group_owner_id: string
        @param src_security_group_owner_id: The ID of the owner of the security group you are 
                                            granting access to.
                                            
        @type ip_protocol: string
        @param ip_protocol: Either tcp | udp | icmp

        @type from_port: int
        @param from_port: The beginning port number you are enabling

        @type to_port: int
        @param to_port: The ending port number you are enabling

        @type to_port: string
        @param to_port: The CIDR block you are providing access to.
                        See http://en.wikipedia.org/wiki/Classless_Inter-Domain_Routing

        @rtype: bool
        @return: True if successful.
        """
        params = {'GroupName':group_name}
        if src_security_group_name:
            params['SourceSecurityGroupName'] = src_security_group_name
        if src_security_group_owner_id:
            params['SourceSecurityGroupOwnerId'] = src_security_group_owner_id
        if ip_protocol:
            params['IpProtocol'] = ip_protocol
        if from_port:
            params['FromPort'] = from_port
        if to_port:
            params['ToPort'] = to_port
        if cidr_ip:
            params['CidrIp'] = urllib.quote(cidr_ip)
        return self.get_status('AuthorizeSecurityGroupIngress', params)

    def revoke_security_group(self, group_name, src_security_group_name=None,
                              src_security_group_owner_id=None,
                              ip_protocol=None, from_port=None, to_port=None,
                              cidr_ip=None):
        """
        Remove an existing rule from an existing security group.
        You need to pass in either src_security_group_name and
        src_security_group_owner_id OR ip_protocol, from_port, to_port,
        and cidr_ip.  In other words, either you are revoking another
        group or you are revoking some ip-based rule.
        
        @type group_name: string
        @param group_name: The name of the security group you are removing
                           the rule from.
                           
        @type src_security_group_name: string
        @param src_security_group_name: The name of the security group you are 
                                        revoking access to.
                                        
        @type src_security_group_owner_id: string
        @param src_security_group_owner_id: The ID of the owner of the security group you are 
                                            revoking access to.
                                            
        @type ip_protocol: string
        @param ip_protocol: Either tcp | udp | icmp

        @type from_port: int
        @param from_port: The beginning port number you are disabling

        @type to_port: int
        @param to_port: The ending port number you are disabling

        @type to_port: string
        @param to_port: The CIDR block you are revoking access to.
                        See http://en.wikipedia.org/wiki/Classless_Inter-Domain_Routing
                        
        @rtype: bool
        @return: True if successful.
        """
        params = {'GroupName':group_name}
        if src_security_group_name:
            params['SourceSecurityGroupName'] = src_security_group_name
        if src_security_group_owner_id:
            params['SourceSecurityGroupOwnerId'] = src_security_group_owner_id
        if ip_protocol:
            params['IpProtocol'] = ip_protocol
        if from_port:
            params['FromPort'] = from_port
        if to_port:
            params['ToPort'] = to_port
        if cidr_ip:
            params['CidrIp'] = cidr_ip
        return self.get_status('RevokeSecurityGroupIngress', params)

    #
    # Regions
    #

    def get_all_regions(self):
        """
        Get all available regions for the EC2 service.
        
        @rtype: list
        @return: A list of L{RegionInfo<boto.ec2.regioninfo.RegionInfo>}
        """
        return self.get_list('DescribeRegions', None, [('item', RegionInfo)])





# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

#
# Parts of this code were copied or derived from sample code supplied by AWS.
# The following notice applies to that code.
#
#  This software code is made available "AS IS" without warranties of any
#  kind.  You may copy, display, modify and redistribute the software
#  code either by itself or as incorporated into your code; provided that
#  you do not remove any proprietary notices.  Your use of this software
#  code is at your own risk and you waive any claim against Amazon
#  Digital Services, Inc. or its affiliates with respect to your use of
#  this software code. (c) 2006 Amazon Digital Services, Inc. or its
#  affiliates.

"""
Handles basic connections to AWS
"""

import base64
import hmac
import httplib
import socket, errno
import re
import sys
import time
import urllib, urlparse
import os
import xml.sax
import boto
from boto.exception import AWSConnectionError, BotoClientError, BotoServerError
from boto.resultset import ResultSet
import boto.utils
from boto import config, UserAgent, handler
try:
    from hashlib import sha1 as sha
    from hashlib import sha256 as sha256
except ImportError:
    import sha
    sha256 = None

PORTS_BY_SECURITY = { True: 443, False: 80 }

class AWSAuthConnection:
    def __init__(self, server, aws_access_key_id=None,
                 aws_secret_access_key=None, is_secure=True, port=None,
                 proxy=None, proxy_port=None, proxy_user=None,
                 proxy_pass=None, debug=0, https_connection_factory=None):
        """
        @type server: string
        @param server: The server to make the connection to
        
        @type aws_access_key_id: string
        @param aws_access_key_id: AWS Access Key ID (provided by Amazon)
        
        @type aws_secret_access_key: string
        @param aws_secret_access_key: Secret Access Key (provided by Amazon)
        
        @type is_secure: boolean
        @param is_secure: Whether the connection is over SSL
        
        @type https_connection_factory: list or tuple
        @param https_connection_factory: A pair of an HTTP connection
                                         factory and the exceptions to catch.
                                         The factory should have a similar
                                         interface to L{httplib.HTTPSConnection}.
        
        @type proxy:
        @param proxy:
        
        @type proxy_port: int
        @param proxy_port: The port to use when connecting over a proxy
        
        @type proxy_user: string
        @param proxy_user: The username to connect with on the proxy
        
        @type proxy_pass: string
        @param proxy_pass: The password to use when connection over a proxy.
        
        @type port: integer
        @param port: The port to use to connect
        """
        
        self.num_retries = 5
        self.is_secure = is_secure
        self.handle_proxy(proxy, proxy_port, proxy_user, proxy_pass)
        # define exceptions from httplib that we want to catch and retry
        self.http_exceptions = (httplib.HTTPException, socket.error, socket.gaierror)
        # define values in socket exceptions we don't want to catch
        self.socket_exception_values = (errno.EINTR,)
        if https_connection_factory is not None:
            self.https_connection_factory = https_connection_factory[0]
            self.http_exceptions += https_connection_factory[1]
        else:
            self.https_connection_factory = None
        if (is_secure):
            self.protocol = 'https'
        else:
            self.protocol = 'http'
        self.server = server
        if debug:
            self.debug = debug
        else:
            self.debug = config.getint('Boto', 'debug', debug)
        if port:
            self.port = port
        else:
            self.port = PORTS_BY_SECURITY[is_secure]
        if self.port == 80:
            self.server_name = server
        else:
            self.server_name = '%s:%d' % (server, self.port)
        
        if aws_access_key_id:
            self.aws_access_key_id = aws_access_key_id
        elif os.environ.has_key('AWS_ACCESS_KEY_ID'):
            self.aws_access_key_id = os.environ['AWS_ACCESS_KEY_ID']
        elif config.has_option('Credentials', 'aws_access_key_id'):
            self.aws_access_key_id = config.get('Credentials', 'aws_access_key_id')
        
        if aws_secret_access_key:
            self.aws_secret_access_key = aws_secret_access_key
        elif os.environ.has_key('AWS_SECRET_ACCESS_KEY'):
            self.aws_secret_access_key = os.environ['AWS_SECRET_ACCESS_KEY']
        elif config.has_option('Credentials', 'aws_secret_access_key'):
            self.aws_secret_access_key = config.get('Credentials', 'aws_secret_access_key')

        # initialize an HMAC for signatures, make copies with each request
        self.hmac = hmac.new(self.aws_secret_access_key, digestmod=sha)
        if sha256:
            self.hmac_256 = hmac.new(self.aws_secret_access_key, digestmod=sha256)
        else:
            self.hmac_256 = None

        # cache up to 20 connections
        self._cache = boto.utils.LRUCache(20)
        self.refresh_http_connection(self.server, self.is_secure)
        self._last_rs = None

    def handle_proxy(self, proxy, proxy_port, proxy_user, proxy_pass):
        self.proxy = proxy
        self.proxy_port = proxy_port
        self.proxy_user = proxy_user
        self.proxy_pass = proxy_pass
        if os.environ.has_key('http_proxy') and not self.proxy:
            pattern = re.compile(
                '(?:http://)?' \
                '(?:(?P<user>\w+):(?P<pass>.*)@)?' \
                '(?P<host>[\w\-\.]+)' \
                '(?::(?P<port>\d+))?'
            )
            match = pattern.match(os.environ['http_proxy'])
            if match:
                self.proxy = match.group('host')
                self.proxy_port = match.group('port')
                self.proxy_user = match.group('user')
                self.proxy_pass = match.group('pass')
        else:
            if not self.proxy:
                self.proxy = config.get_value('Boto', 'proxy', None)
            if not self.proxy_port:
                self.proxy_port = config.get_value('Boto', 'proxy_port', None)
            if not self.proxy_user:
                self.proxy_user = config.get_value('Boto', 'proxy_user', None)
            if not self.proxy_pass:
                self.proxy_pass = config.get_value('Boto', 'proxy_pass', None)

        if not self.proxy_port and self.proxy:
            print "http_proxy environment variable does not specify " \
                "a port, using default"
            self.proxy_port = self.port
        self.use_proxy = (self.proxy != None)
        if self.use_proxy and self.is_secure:
            raise AWSConnectionError("Unable to provide secure connection through proxy")

    def get_http_connection(self, host, is_secure):
        if host is None:
            host = self.server_name
        cached_name = is_secure and 'https://' or 'http://'
        cached_name += host
        if cached_name in self._cache:
            return self._cache[cached_name]
        return self.refresh_http_connection(host, is_secure)

    def refresh_http_connection(self, host, is_secure):
        if self.use_proxy:
            host = '%s:%d' % (self.proxy, int(self.proxy_port))
        if host is None:
            host = self.server_name
        boto.log.debug('establishing HTTP connection')
        if is_secure:
            if self.https_connection_factory:
                connection = self.https_connection_factory(host)
            else:
                connection = httplib.HTTPSConnection(host)
        else:
            connection = httplib.HTTPConnection(host)
        if self.debug > 1:
            connection.set_debuglevel(self.debug)
        cached_name = is_secure and 'https://' or 'http://'
        cached_name += host
        if cached_name in self._cache:
            boto.log.debug('closing old HTTP connection')
            self._cache[cached_name].close()
        self._cache[cached_name] = connection
        # update self.connection for backwards-compatibility
        if host.split(':')[0] == self.server and is_secure == self.is_secure:
            self.connection = connection
        return connection

    def prefix_proxy_to_path(self, path, host=None):
        path = self.protocol + '://' + (host or self.server) + path
        return path

    def get_proxy_auth_header(self):
        auth = base64.encodestring(self.proxy_user+':'+self.proxy_pass)
        return {'Proxy-Authorization': 'Basic %s' % auth}

    def _mexe(self, method, path, data, headers, host=None, sender=None):
        """
        mexe - Multi-execute inside a loop, retrying multiple times to handle
               transient Internet errors by simply trying again.
               Also handles redirects.

        This code was inspired by the S3Utils classes posted to the boto-users
        Google group by Larry Bates.  Thanks!
        """
        boto.log.debug('Method: %s' % method)
        boto.log.debug('Path: %s' % path)
        boto.log.debug('Data: %s' % data)
        boto.log.debug('Headers: %s' % headers)
        boto.log.debug('Host: %s' % host)
        response = None
        body = None
        e = None
        num_retries = config.getint('Boto', 'num_retries', self.num_retries)
        i = 0
        connection = self.get_http_connection(host, self.is_secure)
        while i <= num_retries:
            try:
                if callable(sender):
                    response = sender(connection, method, path, data, headers)
                else:
                    connection.request(method, path, data, headers)
                    response = connection.getresponse()
                location = response.getheader('location')
                # -- gross hack --
                # httplib gets confused with chunked responses to HEAD requests
                # so I have to fake it out
                if method == 'HEAD' and response.chunked:
                    response.chunked = 0
                if response.status == 500 or response.status == 503:
                    boto.log.debug('received %d response, retrying in %d seconds' % (response.status, 2**i))
                    body = response.read()
                elif response.status == 408:
                    body = response.read()
                    print '-------------------------'
                    print '         4 0 8           '
                    print 'path=%s' % path
                    print body
                    print '-------------------------'
                elif response.status < 300 or response.status >= 400 or \
                        not location:
                    return response
                else:
                    scheme, host, path, params, query, fragment = \
                            urlparse.urlparse(location)
                    if query:
                        path += '?' + query
                    boto.log.debug('Redirecting: %s' % scheme + '://' + host + path)
                    connection = self.get_http_connection(host,
                            scheme == 'https')
                    continue
            except KeyboardInterrupt:
                sys.exit('Keyboard Interrupt')
            except self.http_exceptions, e:
                boto.log.debug('encountered %s exception, reconnecting' % \
                                  e.__class__.__name__)
                connection = self.refresh_http_connection(host, self.is_secure)
            time.sleep(2**i)
            i += 1
        # If we made it here, it's because we have exhausted our retries and stil haven't
        # succeeded.  So, if we have a response object, use it to raise an exception.
        # Otherwise, raise the exception that must have already happened.
        if response:
            raise BotoServerError(response.status, response.reason, body)
        elif e:
            raise e
        else:
            raise BotoClientError('Please report this exception as a Boto Issue!')

    def make_request(self, method, path, headers=None, data='', host=None,
            auth_path=None, sender=None):
        if headers == None:
            headers = {'User-Agent' : UserAgent}
        else:
            headers = headers.copy()
        if not headers.has_key('Content-Length'):
            headers['Content-Length'] = len(data)
        if self.use_proxy:
            path = self.prefix_proxy_to_path(path, host)
            if self.proxy_user and self.proxy_pass:
                headers.update(self.get_proxy_auth_header())
        self.add_aws_auth_header(headers, method, auth_path or path)
        return self._mexe(method, path, data, headers, host, sender)

    def add_aws_auth_header(self, headers, method, path):
        if not headers.has_key('Date'):
            headers['Date'] = time.strftime("%a, %d %b %Y %H:%M:%S GMT",
                                            time.gmtime())

        c_string = boto.utils.canonical_string(method, path, headers)
        boto.log.debug('Canonical: %s' % c_string)
        hmac = self.hmac.copy()
        hmac.update(c_string)
        b64_hmac = base64.encodestring(hmac.digest()).strip()
        headers['Authorization'] = "AWS %s:%s" % (self.aws_access_key_id, b64_hmac)

class AWSQueryConnection(AWSAuthConnection):

    APIVersion = ''
    SignatureVersion = '1'
    ResponseError = BotoServerError

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                 is_secure=True, port=None, proxy=None, proxy_port=None,
                 proxy_user=None, proxy_pass=None, host=None, debug=0,
                 https_connection_factory=None):
        AWSAuthConnection.__init__(self, host, aws_access_key_id, aws_secret_access_key,
                                   is_secure, port, proxy, proxy_port, proxy_user, proxy_pass,
                                   debug,  https_connection_factory)

    def get_utf8_value(self, value):
        if not isinstance(value, str) and not isinstance(value, unicode):
            value = str(value)
        if isinstance(value, unicode):
            return value.encode('utf-8')
        else:
            return unicode(value, encoding='utf-8')

    def calc_signature_0(self, params):
        boto.log.debug('using calc_signature_0')
        hmac = self.hmac.copy()
        s = params['Action'] + params['Timestamp']
        hmac.update(s)
        keys = params.keys()
        keys.sort(cmp = lambda x, y: cmp(x.lower(), y.lower()))
        pairs = []
        for key in keys:
            val = self.get_utf8_value(params[key])
            pairs.append(key + '=' + urllib.quote(val))
        qs = '&'.join(pairs)
        return (qs, base64.b64encode(hmac.digest()))

    def calc_signature_1(self, params):
        boto.log.debug('using calc_signature_1')
        hmac = self.hmac.copy()
        keys = params.keys()
        keys.sort(cmp = lambda x, y: cmp(x.lower(), y.lower()))
        pairs = []
        for key in keys:
            hmac.update(key)
            val = self.get_utf8_value(params[key])
            hmac.update(val)
            pairs.append(key + '=' + urllib.quote(val))
        qs = '&'.join(pairs)
        return (qs, base64.b64encode(hmac.digest()))

    def calc_signature_2(self, params, verb, path):
        boto.log.debug('using calc_signature_2')
        string_to_sign = '%s\n%s\n%s\n' % (verb, self.server_name.lower(),
                                           path or '/')
        if self.hmac_256:
            hmac = self.hmac_256.copy()
            params['SignatureMethod'] = 'HmacSHA256'
        else:
            hmac = self.hmac.copy()
            params['SignatureMethod'] = 'HmacSHA1'
        keys = params.keys()
        keys.sort()
        pairs = []
        for key in keys:
            val = self.get_utf8_value(params[key])
            pairs.append(urllib.quote(key, safe='') + '=' + urllib.quote(val, safe=''))
        qs = '&'.join(pairs)
        boto.log.debug('query string: %s' % qs)
        string_to_sign += qs
        boto.log.debug('string_to_sign: %s' % string_to_sign)
        hmac.update(string_to_sign)
        return (qs, base64.b64encode(hmac.digest()))

    def get_signature(self, params, verb, path):
        if self.SignatureVersion == '0':
            t = self.calc_signature_0(params)
        elif self.SignatureVersion == '1':
            t = self.calc_signature_1(params)
        elif self.SignatureVersion == '2':
            t = self.calc_signature_2(params, verb, path)
        else:
            raise BotoClientError('Unknown Signature Version: %s' % self.SignatureVersion)
        return t

    def make_request(self, action, params=None, path=None, verb='GET'):
        headers = {'User-Agent' : UserAgent}
        if path == None:
            path = '/'
        if params == None:
            params = {}
        params['Action'] = action
        params['Version'] = self.APIVersion
        params['AWSAccessKeyId'] = self.aws_access_key_id
        params['SignatureVersion'] = self.SignatureVersion
        params['Timestamp'] = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime())
        qs, signature = self.get_signature(params, verb, path)
        qs = path + '?' + qs + '&Signature=' + urllib.quote(signature)
        if self.use_proxy:
            qs = self.prefix_proxy_to_path(qs)
        return self._mexe(verb, qs, None, headers)

    def build_list_params(self, params, items, label):
        if isinstance(items, str):
            items = [items]
        for i in range(1, len(items)+1):
            params['%s.%d' % (label, i)] = items[i-1]
            
    # generics

    def get_list(self, action, params, markers, path='/', parent=None):
        if not parent:
            parent = self
        response = self.make_request(action, params, path)
        body = response.read()
        if response.status == 200:
            rs = ResultSet(markers)
            h = handler.XmlHandler(rs, parent)
            xml.sax.parseString(body, h)
            return rs
        else:
            boto.log.error('%s %s' % (response.status, response.reason))
            boto.log.error('%s' % body)
            raise self.ResponseError(response.status, response.reason, body)
        
    def get_object(self, action, params, cls, path='/', parent=None):
        if not parent:
            parent = self
        response = self.make_request(action, params, path)
        body = response.read()
        if response.status == 200:
            obj = cls(parent)
            h = handler.XmlHandler(obj, parent)
            xml.sax.parseString(body, h)
            return obj
        else:
            boto.log.error('%s %s' % (response.status, response.reason))
            boto.log.error('%s' % body)
            raise self.ResponseError(response.status, response.reason, body)
        
    def get_status(self, action, params, path='/', parent=None):
        if not parent:
            parent = self
        response = self.make_request(action, params, path)
        body = response.read()
        if response.status == 200:
            rs = ResultSet()
            h = handler.XmlHandler(rs, parent)
            xml.sax.parseString(body, h)
            return rs.status
        else:
            boto.log.error('%s %s' % (response.status, response.reason))
            boto.log.error('%s' % body)
            raise self.ResponseError(response.status, response.reason, body)




# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Represents an SDB Domain
"""
from boto.sdb.queryresultset import QueryResultSet
from boto.sdb.item import Item

class Domain:
    
    def __init__(self, connection=None, name=None):
        self.connection = connection
        self.name = name
        self._metadata = None

    def __repr__(self):
        return 'Domain:%s' % self.name

    def __iter__(self):
        return iter(QueryResultSet(self))

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'DomainName':
            self.name = value
        else:
            setattr(self, name, value)

    def get_metadata(self):
        if not self._metadata:
            self._metadata = self.connection.domain_metadata(self)
        return self._metadata
    
    def put_attributes(self, item_name, attributes, replace=True):
        return self.connection.put_attributes(self, item_name, attributes, replace)

    def get_attributes(self, item_name, attribute_name=None, item=None):
        return self.connection.get_attributes(self, item_name, attribute_name, item)

    def delete_attributes(self, item_name, attributes=None):
        return self.connection.delete_attributes(self, item_name, attributes)

    def query(self, query='', max_items=None, attr_names=None):
        """
        Returns a list of items within domain that match the query.
        
        @type query: string
        @param query: The SimpleDB query to be performed.

        @type max_items: int
        @param max_items: The maximum number of items to return.  If not
                          supplied, the default is None which returns all
                          items matching the query.

        @type attr_names: list
        @param attr_names: Either None, meaning return all attributes
                           or a list of attribute names which means to return
                           only those attributes.

        @rtype: iter
        @return: An iterator containing the results.  This is actually a generator
                 function that will iterate across all search results, not just the
                 first page.
        """
        return iter(QueryResultSet(self, query, max_items, attr_names))
    
    def get_item(self, item_name):
        item = self.get_attributes(item_name)
        if item:
            item.domain = self
            return item
        else:
            return None

    def new_item(self, item_name):
        return Item(self, item_name)

    def delete_item(self, item):
        self.delete_attributes(item.name)

class DomainMetaData:
    
    def __init__(self, domain=None):
        self.domain = domain
        self.item_count = None
        self.item_names_size = None
        self.attr_name_count = None
        self.attr_names_size = None
        self.attr_value_count = None
        self.attr_values_size = None

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'ItemCount':
            self.item_count = int(value)
        elif name == 'ItemNamesSizeBytes':
            self.item_names_size = int(value)
        elif name == 'AttributeNameCount':
            self.attr_name_count = int(value)
        elif name == 'AttributeNamesSizeBytes':
            self.attr_names_size = int(value)
        elif name == 'AttributeValueCount':
            self.attr_value_count = int(value)
        elif name == 'AttributeValuesSizeBytes':
            self.attr_values_size = int(value)
        elif name == 'Timestamp':
            self.timestamp = value
        else:
            setattr(self, name, value)




# Copyright (c) 2008 Chris Moyer http://coredumped.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

class Qualifications:

    def __init__(self, requirements = []):
        self.requirements = requirements

    def add(self, req):
        self.requirements.append(req)

    def get_as_params(self):
        params = {}
        assert(len(self.requirements) <= 10)
        for n, req in enumerate(self.requirements):
            params["QualificationRequirement.%s.QualificationTypeId" % (n+1)] = req.qualification_type_id
            params["QualificationRequirement.%s.Comparator" % (n+1)] = req.comparator
            params["QualificationRequirement.%s.IntegerValue" % (n+1)] = req.integer_value
            if req.required_to_preview:
                params["QualificationRequirement.%s.RequiredToPreview" % (n+1)] = "true"
        return params


class Requirement:
    """
    Representation of a single requirement
    """

    def __init__(self, qualification_type_id, comparator, integer_value, required_to_preview=False):
        self.qualification_type_id = qualification_type_id
        self.comparator = comparator
        self.integer_value = integer_value
        self.required_to_preview = required_to_preview

class PercentAssignmentsSubmittedRequirement(Requirement):
    """
    The percentage of assignments the Worker has submitted, over all assignments the Worker has accepted. The value is an integer between 0 and 100.
    """

    def __init__(self, comparator, integer_value, required_to_preview=False):
        Requirement.__init__(self, qualification_type_id="00000000000000000000", comparator=comparator, integer_value=integer_value, required_to_preview=required_to_preview)

class PercentAssignmentsAbandonedRequirement(Requirement):
    """
    The percentage of assignments the Worker has abandoned (allowed the deadline to elapse), over all assignments the Worker has accepted. The value is an integer between 0 and 100.
    """

    def __init__(self, comparator, integer_value, required_to_preview=False):
        Requirement.__init__(self, qualification_type_id="00000000000000000070", comparator=comparator, integer_value=integer_value, required_to_preview=required_to_preview)

class PercentAssignmentsReturnedRequirement(Requirement):
    """
    The percentage of assignments the Worker has returned, over all assignments the Worker has accepted. The value is an integer between 0 and 100.
    """

    def __init__(self, comparator, integer_value, required_to_preview=False):
        Requirement.__init__(self, qualification_type_id="000000000000000000E0", comparator=comparator, integer_value=integer_value, required_to_preview=required_to_preview)

class PercentAssignmentsApprovedRequirement(Requirement):
    """
    The percentage of assignments the Worker has submitted that were subsequently approved by the Requester, over all assignments the Worker has submitted. The value is an integer between 0 and 100.
    """

    def __init__(self, comparator, integer_value, required_to_preview=False):
        Requirement.__init__(self, qualification_type_id="000000000000000000L0", comparator=comparator, integer_value=integer_value, required_to_preview=required_to_preview)

class PercentAssignmentsRejectedRequirement(Requirement):
    """
    The percentage of assignments the Worker has submitted that were subsequently rejected by the Requester, over all assignments the Worker has submitted. The value is an integer between 0 and 100.
    """

    def __init__(self, comparator, integer_value, required_to_preview=False):
        Requirement.__init__(self, qualification_type_id="000000000000000000S0", comparator=comparator, integer_value=integer_value, required_to_preview=required_to_preview)



# Copyright (c) 2006,2007 Mitch Garnaat http://garnaat.org/
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import urllib
import xml.sax
import threading
import boto
from boto import handler
from boto.connection import AWSQueryConnection
from boto.sdb.domain import Domain, DomainMetaData
from boto.sdb.item import Item
from boto.exception import SDBResponseError
from boto.resultset import ResultSet

class ItemThread(threading.Thread):
    
    def __init__(self, name, domain_name, item_names):
        threading.Thread.__init__(self, name=name)
        print 'starting %s with %d items' % (name, len(item_names))
        self.domain_name = domain_name
        self.conn = SDBConnection()
        self.item_names = item_names
        self.items = []
        
    def run(self):
        for item_name in self.item_names:
            item = self.conn.get_attributes(self.domain_name, item_name)
            self.items.append(item)

class SDBConnection(AWSQueryConnection):

    APIVersion = '2007-11-07'
    SignatureVersion = '2'
    ResponseError = SDBResponseError

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                 is_secure=True, port=None, proxy=None, proxy_port=None,
                 proxy_user=None, proxy_pass=None, host='sdb.amazonaws.com', debug=0,
                 https_connection_factory=None):
        AWSQueryConnection.__init__(self, aws_access_key_id, aws_secret_access_key,
                                    is_secure, port, proxy, proxy_port, proxy_user, proxy_pass,
                                    host, debug, https_connection_factory)
        self.box_usage = 0.0

    def build_name_value_list(self, params, attributes, replace=False):
        keys = attributes.keys()
        keys.sort()
        i = 1
        for key in keys:
            value = attributes[key]
            if isinstance(value, list):
                for v in value:
                    params['Attribute.%d.Name'%i] = key
                    params['Attribute.%d.Value'%i] = v
                    if replace:
                        params['Attribute.%d.Replace'%i] = 'true'
                    i += 1
            else:
                params['Attribute.%d.Name'%i] = key
                params['Attribute.%d.Value'%i] = value
                if replace:
                    params['Attribute.%d.Replace'%i] = 'true'
            i += 1

    def build_name_list(self, params, attribute_names):
        i = 1
        attribute_names.sort()
        for name in attribute_names:
            params['Attribute.%d.Name'%i] = name
            i += 1

    def get_usage(self):
        """
        Returns the BoxUsage accumulated on this SDBConnection object.

        @rtype: float
        @return: The accumulated BoxUsage of all requests made on the connection.
        """
        return self.box_usage

    def print_usage(self):
        """
        Print the BoxUsage and approximate costs of all requests made on this connection.
        """
        print 'Total Usage: %f compute seconds' % self.box_usage
        cost = self.box_usage * 0.14
        print 'Approximate Cost: $%f' % cost

    def get_domain(self, domain_name, validate=True):
        domain = Domain(self, domain_name)
        if validate:
            self.query(domain, '', max_items=1)
        return domain

    def lookup(self, domain_name, validate=True):
        """
        Lookup an existing SimpleDB domain

        @type domain_name: string
        @param domain_name: The name of the new domain

        @rtype: L{Domain<boto.sdb.domain.Domain>} object or None
        @return: The Domain object or None if the domain does not exist.
        """
        try:
            domain = self.get_domain(domain_name, validate)
        except:
            domain = None
        return domain

    def get_all_domains(self, max_domains=None, next_token=None):
        params = {}
        if max_domains:
            params['MaxNumberOfDomains'] = max_domains
        if next_token:
            params['NextToken'] = next_token
        return self.get_list('ListDomains', params, [('DomainName', Domain)])
        
    def create_domain(self, domain_name):
        """
        Create a SimpleDB domain.

        @type domain_name: string
        @param domain_name: The name of the new domain

        @rtype: L{Domain<boto.sdb.domain.Domain>} object
        @return: The newly created domain
        """
        params = {'DomainName':domain_name}
        d = self.get_object('CreateDomain', params, Domain)
        d.name = domain_name
        return d

    def get_domain_and_name(self, domain_or_name):
        if (isinstance(domain_or_name, Domain)):
            return (domain_or_name, domain_or_name.name)
        else:
            return (self.get_domain(domain_or_name), domain_or_name)
        
    def delete_domain(self, domain_or_name):
        """
        Delete a SimpleDB domain.

        @type domain_or_name: string or L{Domain<boto.sdb.domain.Domain>} object.
        @param domain_or_name: Either the name of a domain or a Domain object

        @rtype: bool
        @return: True if successful
        
        B{Note:} This will delete the domain and all items within the domain.
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        params = {'DomainName':domain_name}
        return self.get_status('DeleteDomain', params)
        
    def domain_metadata(self, domain_or_name):
        """
        Get the Metadata for a SimpleDB domain.

        @type domain_or_name: string or L{Domain<boto.sdb.domain.Domain>} object.
        @param domain_or_name: Either the name of a domain or a Domain object

        @rtype: L{DomainMetaData<boto.sdb.domain.DomainMetaData>} object
        @return: The newly created domain metadata object
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        params = {'DomainName':domain_name}
        d = self.get_object('DomainMetadata', params, DomainMetaData)
        d.domain = domain
        return d
        
    def put_attributes(self, domain_or_name, item_name, attributes, replace=True):
        """
        Store attributes for a given item in a domain.

        @type domain_or_name: string or L{Domain<boto.sdb.domain.Domain>} object.
        @param domain_or_name: Either the name of a domain or a Domain object

        @type item_name: string
        @param item_name: The name of the item whose attributes are being stored.

        @type attribute_names: dict or dict-like object
        @param attribute_names: The name/value pairs to store as attributes

        @type replace: bool
        @param replace: Whether the attribute values passed in will replace
                        existing values or will be added as addition values.
                        Defaults to True.

        @rtype: bool
        @return: True if successful
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        params = {'DomainName' : domain_name,
                  'ItemName' : item_name}
        self.build_name_value_list(params, attributes, replace)
        return self.get_status('PutAttributes', params)

    def get_attributes(self, domain_or_name, item_name, attribute_names=None, item=None):
        """
        Retrieve attributes for a given item in a domain.

        @type domain_or_name: string or L{Domain<boto.sdb.domain.Domain>} object.
        @param domain_or_name: Either the name of a domain or a Domain object

        @type item_name: string
        @param item_name: The name of the item whose attributes are being retrieved.

        @type attribute_names: string or list of strings
        @param attribute_names: An attribute name or list of attribute names.  This
                                parameter is optional.  If not supplied, all attributes
                                will be retrieved for the item.

        @rtype: L{Item<boto.sdb.item.Item>}
        @return: An Item mapping type containing the requested attribute name/values
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        params = {'DomainName' : domain_name,
                  'ItemName' : item_name}
        if attribute_names:
            if not isinstance(attribute_names, list):
                attribute_names = [attribute_names]
            self.build_list_params(params, attribute_names, 'AttributeName')
        response = self.make_request('GetAttributes', params)
        body = response.read()
        if response.status == 200:
            if item == None:
                item = Item(domain, item_name)
            h = handler.XmlHandler(item, self)
            xml.sax.parseString(body, h)
            return item
        else:
            raise SDBResponseError(response.status, response.reason, body)
        
    def delete_attributes(self, domain_or_name, item_name, attr_names=None):
        """
        Delete attributes from a given item in a domain.

        @type domain_or_name: string or L{Domain<boto.sdb.domain.Domain>} object.
        @param domain_or_name: Either the name of a domain or a Domain object

        @type item_name: string
        @param item_name: The name of the item whose attributes are being deleted.

        @type attributes: dict, list or L{Item<boto.sdb.item.Item>}
        @param attributes: Either a list containing attribute names which will cause
                           all values associated with that attribute name to be deleted or
                           a dict or Item containing the attribute names and keys and list
                           of values to delete as the value.  If no value is supplied,
                           all attribute name/values for the item will be deleted.
                           
        @rtype: bool
        @return: True if successful
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        params = {'DomainName':domain_name,
                  'ItemName' : item_name}
        if attr_names:
            if isinstance(attr_names, list):
                self.build_name_list(params, attr_names)
            elif isinstance(attr_names, dict) or isinstance(attr_names, Item):
                self.build_name_value_list(params, attr_names)
        return self.get_status('DeleteAttributes', params)
        
    def query(self, domain_or_name, query='', max_items=None, next_token=None):
        """
        Returns a list of item names within domain_name that match the query.
        
        @type domain_or_name: string or L{Domain<boto.sdb.domain.Domain>} object.
        @param domain_or_name: Either the name of a domain or a Domain object

        @type query: string
        @param query: The SimpleDB query to be performed.

        @type max_items: int
        @param max_items: The maximum number of items to return.  If not
                          supplied, the default is None which returns all
                          items matching the query.

        @rtype: ResultSet
        @return: An iterator containing the results.
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        params = {'DomainName':domain_name,
                  'QueryExpression' : query}
        if max_items:
            params['MaxNumberOfItems'] = max_items
        if next_token:
            params['NextToken'] = next_token
        return self.get_object('Query', params, ResultSet)

    def query_with_attributes(self, domain_or_name, query='', attr_names=None,
                              max_items=None, next_token=None):
        """
        Returns a set of Attributes for item names within domain_name that match the query.
        
        @type domain_or_name: string or L{Domain<boto.sdb.domain.Domain>} object.
        @param domain_or_name: Either the name of a domain or a Domain object

        @type query: string
        @param query: The SimpleDB query to be performed.

        @type attr_names: list
        @param attr_names: The name of the attributes to be returned.
                           If no attributes are specified, all attributes
                           will be returned.

        @type max_items: int
        @param max_items: The maximum number of items to return.  If not
                          supplied, the default is None which returns all
                          items matching the query.

        @rtype: ResultSet
        @return: An iterator containing the results.
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        params = {'DomainName':domain_name,
                  'QueryExpression' : query}
        if max_items:
            params['MaxNumberOfItems'] = max_items
        if next_token:
            params['NextToken'] = next_token
        if attr_names:
            self.build_list_params(params, attr_names, 'AttributeName')
        return self.get_list('QueryWithAttributes', params, [('Item', Item)], parent=domain)

    def threaded_query(self, domain_or_name, query='', max_items=None, next_token=None, num_threads=6):
        """
        Returns a list of fully populated items that match the query provided.

        The name/value pairs for all of the matching item names are retrieved in a number of separate
        threads (specified by num_threads) to achieve maximum throughput.
        The ResultSet that is returned has an attribute called next_token that can be used
        to retrieve additional results for the same query.
        """
        domain, domain_name = self.get_domain_and_name(domain_or_name)
        if max_items and num_threads > max_items:
            num_threads = max_items
        rs = self.query(domain_or_name, query, max_items, next_token)
        threads = []
        n = len(rs) / num_threads
        for i in range(0, num_threads):
            if i+1 == num_threads:
                thread = ItemThread('Thread-%d' % i, domain_name, rs[n*i:])
            else:
                thread = ItemThread('Thread-%d' % i, domain_name, rs[n*i:n*(i+1)])
            threads.append(thread)
            thread.start()
        del rs[0:]
        for thread in threads:
            thread.join()
            for item in thread.items:
                rs.append(item)
        return rs




from boto.sdb.persist.object import SDBObject
from boto.sdb.persist.property import *
from boto.sdb.persist import Manager
from datetime import datetime
import time

#
# This will eventually be moved to the boto.tests module and become a real unit test
# but for now it will live here.  It shows examples of each of the Property types in
# use and tests the basic operations.
#
class TestScalar(SDBObject):

    name = StringProperty()
    description = StringProperty()
    size = PositiveIntegerProperty()
    offset = IntegerProperty()
    foo = BooleanProperty()
    date = DateTimeProperty()
    file = S3KeyProperty()

class TestRef(SDBObject):

    name = StringProperty()
    ref = ObjectProperty(ref_class=TestScalar)

class TestSubClass1(TestRef):

    answer = PositiveIntegerProperty()

class TestSubClass2(TestScalar):

    flag = BooleanProperty()

class TestList(SDBObject):

    names = StringListProperty()
    numbers = PositiveIntegerListProperty()
    bools = BooleanListProperty()
    objects = ObjectListProperty(ref_class=TestScalar)
    
def test1():
    s = TestScalar()
    s.name = 'foo'
    s.description = 'This is foo'
    s.size = 42
    s.offset = -100
    s.foo = True
    s.date = datetime.now()
    s.save()
    return s

def test2(ref_name):
    s = TestRef()
    s.name = 'testref'
    rs = TestScalar.find(name=ref_name)
    s.ref = rs.next()
    s.save()
    return s

def test3():
    s = TestScalar()
    s.name = 'bar'
    s.description = 'This is bar'
    s.size = 24
    s.foo = False
    s.date = datetime.now()
    s.save()
    return s

def test4(ref1, ref2):
    s = TestList()
    s.names.append(ref1.name)
    s.names.append(ref2.name)
    s.numbers.append(ref1.size)
    s.numbers.append(ref2.size)
    s.bools.append(ref1.foo)
    s.bools.append(ref2.foo)
    s.objects.append(ref1)
    s.objects.append(ref2)
    s.save()
    return s

def test5(ref):
    s = TestSubClass1()
    s.answer = 42
    s.ref = ref
    s.save()
    # test out free form attribute
    s.fiddlefaddle = 'this is fiddlefaddle'
    s._fiddlefaddle = 'this is not fiddlefaddle'
    return s

def test6():
    s = TestSubClass2()
    s.name = 'fie'
    s.description = 'This is fie'
    s.size = 4200
    s.offset = -820
    s.foo = False
    s.date = datetime.now()
    s.flag = True
    s.save()
    return s

def test(domain_name):
    print 'Initialize the Persistance system'
    Manager.DefaultDomainName = domain_name
    print 'Call test1'
    s1 = test1()
    # now create a new instance and read the saved data from SDB
    print 'Now sleep to wait for things to converge'
    time.sleep(5)
    print 'Now lookup the object and compare the fields'
    s2 = TestScalar(s1.id)
    assert s1.name == s2.name
    assert s1.description == s2.description
    assert s1.size == s2.size
    assert s1.offset == s2.offset
    assert s1.foo == s2.foo
    #assert s1.date == s2.date
    print 'Call test2'
    s2 = test2(s1.name)
    print 'Call test3'
    s3 = test3()
    print 'Call test4'
    s4 = test4(s1, s3)
    print 'Call test5'
    s6 = test6()
    s5 = test5(s6)
    domain = s5._manager.domain
    item1 = domain.get_item(s1.id)
    item2 = domain.get_item(s2.id)
    item3 = domain.get_item(s3.id)
    item4 = domain.get_item(s4.id)
    item5 = domain.get_item(s5.id)
    item6 = domain.get_item(s6.id)
    return [(s1, item1), (s2, item2), (s3, item3), (s4, item4), (s5, item5), (s6, item6)]
